import os
import json
import re
import asyncio
import requests
from urllib.parse import urlparse, urljoin
from functools import partial
import boto3
from requests_aws4auth import AWS4Auth

from biothings import config as btconfig
from biothings.hub.dataload.dumper import HTTPDumper, DumperException
from biothings.utils.common import uncompressall, md5sum


class BiothingsDumper(HTTPDumper):
    """
    This dumper is used to maintain a BioThings API up-to-date. BioThings data
    is available as either as an ElasticSearch snapshot when full update,
    and a collection of diff files for incremental updates.
    It will either download incremental updates and apply diff, or trigger an ElasticSearch
    restore if the latest version is a full update.
    This dumper can also be configured with precedence rules: when a full and a incremental
    update is available, rules can set so full is preferably used over incremental (size can also
    be considered when selecting the preferred way).
    """
    # URL pointing to versions.json file, this is the main entry point
    VERSION_URL = None

    # set during autohub init
    SRC_NAME = None
    SRC_ROOT_FOLDER = None

    # Auto-deploy data update ?
    AUTO_UPLOAD = False

    # Optionally, a schedule can be used to automatically check new version
    #SCHEDULE = "0 9 * * *"

    # what backend the dumper should work with. Must be defined before instantiation
    # (can be an instance or a partial() returning an instance)
    TARGET_BACKEND = None

    # TODO: should we ensure ARCHIVE is always true ?
    # (ie we have to keep all versions to apply them in order)

    # must be set before use if accessing restricted bucket
    AWS_ACCESS_KEY_ID = None
    AWS_SECRET_ACCESS_KEY = None

    def __init__(self, *args, **kwargs):
        super(BiothingsDumper, self).__init__(*args, **kwargs)
        # list of build_version to download/apply, in order
        self._target_backend = None

    def prepare_client(self):
        """
        Depending on presence of credentials, inject authentication
        in client.get()
        """
        super().prepare_client()
        if self.__class__.AWS_ACCESS_KEY_ID and self.__class__.AWS_SECRET_ACCESS_KEY:
            self._client = requests.Session()
            self._client.verify = self.__class__.VERIFY_CERT

            def auth_get(url, *args, **kwargs):
                if ".s3-website-" in url:
                    raise DumperException(
                        "Can't access s3 static website using authentication")
                # extract region from URL (reliable ?)
                pat = re.compile(r"https?://(.*)\.(.*)\.amazonaws.com.*")
                m = pat.match(url)
                if m:
                    bucket_name, frag = m.groups()
                    # looks like "s3-us-west-2"
                    # whether static website is activated or not
                    region = frag.replace("s3-", "")
                    if region == "s3":  # url doesn't contain a region, we need to query the bucket
                        s3client = boto3.client(
                            "s3",
                            aws_access_key_id=self.__class__.AWS_ACCESS_KEY_ID,
                            aws_secret_access_key=self.__class__.
                            AWS_SECRET_ACCESS_KEY)
                        bucket_info = s3client.get_bucket_location(
                            Bucket=bucket_name)
                        region = bucket_info["LocationConstraint"]
                    auth = AWS4Auth(self.__class__.AWS_ACCESS_KEY_ID,
                                    self.__class__.AWS_SECRET_ACCESS_KEY,
                                    region, 's3')
                    return self._client.get(url, auth=auth, *args, **kwargs)
                else:
                    raise DumperException(
                        "Couldn't determine s3 region from url '%s'" % url)

            self.client.get = auth_get

    @property
    def base_url(self):
        # add trailing / so urljoin won't remove folder from path
        return os.path.dirname(self.__class__.VERSION_URL) + "/"

    @property
    def target_backend(self):
        if not self._target_backend:
            if type(self.__class__.TARGET_BACKEND) == partial:
                self._target_backend = self.__class__.TARGET_BACKEND()
            else:
                self._target_backend = self.__class__.TARGET_BACKEND
        return self._target_backend

    @asyncio.coroutine
    def get_target_backend(self):
        @asyncio.coroutine
        def do():
            cnt = self.target_backend.count()
            return {
                "host": self.target_backend.target_esidxer.es_host,
                "index": self.target_backend.target_name,
                "version": self.target_backend.version,
                "count": cnt
            }

        return do()

    @asyncio.coroutine
    def reset_target_backend(self):
        @asyncio.coroutine
        def do():
            if self.target_backend.target_esidxer.exists_index():
                self.target_backend.target_esidxer.delete_index()

        return do()

    def download(self, remoteurl, localfile, headers={}):
        self.prepare_local_folders(localfile)
        parsed = urlparse(remoteurl)
        if self.__class__.AWS_ACCESS_KEY_ID and self.__class__.AWS_SECRET_ACCESS_KEY:
            # accessing diffs controled by auth
            key = parsed.path.strip(
                "/")  # s3 key are relative, not / at beginning
            # extract bucket name from URL (reliable?)
            pat = re.compile(r"^(.*?)\..*\.amazonaws.com")
            m = pat.match(parsed.netloc)
            if m:
                bucket_name = m.groups()[0]
            else:
                raise DumperException(
                    "Can't extract bucket name from URL '%s'" % remote_url)

            return self.auth_download(bucket_name, key, localfile, headers)
        else:
            return self.anonymous_download(remoteurl, localfile, headers)

    def anonymous_download(self, remoteurl, localfile, headers={}):
        res = super(BiothingsDumper, self).download(remoteurl,
                                                    localfile,
                                                    headers=headers)
        # use S3 metadata to set local mtime
        # we add 1 second to make sure we wouldn't download remoteurl again
        # because remote is older by just a few milliseconds
        lastmodified = int(res.headers["x-amz-meta-lastmodified"]) + 1
        os.utime(localfile, (lastmodified, lastmodified))
        return res

    def auth_download(self, bucket_name, key, localfile, headers={}):
        session = boto3.Session(
            aws_access_key_id=self.__class__.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=self.__class__.AWS_SECRET_ACCESS_KEY)
        bucket = session.resource("s3").Bucket(bucket_name)
        res = bucket.download_file(key, localfile)
        return res

    def check_compat(self, build_meta):
        if hasattr(btconfig,
                   "SKIP_CHECK_COMPAT") and btconfig.SKIP_CHECK_COMPAT:
            return

        msg = []
        for version_field in [
                "app_version", "standalone_version", "biothings_version"
        ]:
            VERSION_FIELD = version_field.upper()
            version = build_meta.get(version_field)
            assert version is not None, "Version field '%s' is None" % VERSION_FIELD
            # some releases use dict (most recent) some use string
            if isinstance(version, dict):
                version = version["branch"]
            if type(version) != list:
                version = [version]
            # remove hash from versions (only useful when version is a string,
            # not a dict, see above
            version = [re.sub(r"( \[.*\])", "", v) for v in version]
            version = set(version)
            if version == set([None]):
                raise DumperException(
                    "Remote data is too old and can't be handled with current app (%s not defined)"
                    % version_field)
            versionfromconf = re.sub(
                r"( \[.*\])", "",
                getattr(btconfig, VERSION_FIELD).get("branch"))
            VERSION = set()
            VERSION.add(versionfromconf)
            found_compat_version = VERSION.intersection(version)
            assert found_compat_version, "Remote data requires %s to be %s, but current app is %s" % (
                version_field, version, VERSION)
            msg.append("%s=%s:OK" % (version_field, version))

    def load_remote_json(self, url):
        res = self.client.get(url, allow_redirects=True)
        redirect = res.headers.get('x-amz-website-redirect-location')
        if redirect:
            parsed = urlparse(url)
            newurl = parsed._replace(path=redirect)
            res = self.client.get(newurl.geturl())
        if res.status_code != 200:
            self.logger.error(res)
            return None
        try:
            jsondat = json.loads(res.text)
            return jsondat
        except json.JSONDecodeError:
            self.logger.error(res.headers)
            self.logger.error(res)
            return None

    def compare_remote_local(self, remote_version, local_version,
                             orig_remote_version, orig_local_version):
        # we need have to some data locally. do we already have ?
        if remote_version > local_version:
            self.logger.info("Remote version '%s' is more recent than local version '%s', download needed" %
                             (orig_remote_version, orig_local_version))
            return True
        else:
            self.logger.info("Remote version '%s' is the same as " % orig_remote_version
                             + "local version '%s'. " % orig_local_version + "Dump is waiting to be applied")
            return False

    def remote_is_better(self, remotefile, localfile):
        remote_dat = self.load_remote_json(remotefile)
        if not remote_dat:
            self.logger.info("Couldn't find any build metadata at url '%s'" %
                             remotefile)
            return False
        orig_remote_version = remote_dat["build_version"]

        local_dat = json.load(open(localfile))
        orig_local_version = local_dat["build_version"]

        # if diff version, we want to compatr the right part (destination version)
        # local: "3.4", backend: "4". It's actually the same (4==4)
        local_version = orig_local_version.split(".")[-1]
        remote_version = orig_remote_version.split(".")[-1]
        if remote_version != orig_remote_version:
            self.logger.debug("Remote version '%s' converted to '%s' " % (orig_remote_version, remote_version)
                              + "(version that will be reached once incremental update has been applied)")
        if local_version != orig_local_version:
            self.logger.debug("Local version '%s' converted to '%s' " % (orig_local_version, local_version)
                              + "(version that had been be reached using incremental update files)")

        backend_version = self.target_backend.version
        if backend_version is None:
            self.logger.info("No backend version found")
            return self.compare_remote_local(remote_version, local_version,
                                             orig_remote_version,
                                             orig_local_version)
        elif remote_version > backend_version:
            self.logger.info("Remote version '%s' is more recent than backend version '%s'" %
                             (orig_remote_version, backend_version))
            return self.compare_remote_local(remote_version, local_version,
                                             orig_remote_version,
                                             orig_local_version)
        else:
            self.logger.info("Backend version '%s' is up-to-date" %
                             backend_version)
            return False

    def choose_best_version(self, versions):
        """
        Out of all compatible versions, choose the best:
        1. choose incremental vs. full according to preferences
        2. version must be the highest (most up-to-date)
        """
        # 1st pass
        # TODO: implemente inc/full preferences, for now prefer incremental
        if not versions:
            raise DumperException("No compatible version found")
        preferreds = [v for v in versions if "." in v]
        if preferreds:
            self.logger.info(
                "Preferred versions (according to preferences): %s" %
                preferreds)
            versions = preferreds
        # we can directly take the max because:
        # - version is a string
        # - format if YYYYMMDD
        # - when incremental, it's always old_version.new_version
        return max(versions, key=lambda e: e["build_version"])

    def find_update_path(self, version, backend_version=None):
        """
        Explore available versions and find the path to update the hub up to "version",
        starting from given backend_version (typically current version found in ES index).
        If backend_version is None (typically no index yet), a complete path will be returned,
        from the last compatible "full" release up-to the latest "diff" update.
        Returned is a list of dict, where each dict is a build metadata element containing
        information about each update (see versions.json), the order of the list describes
        the order the updates should be performed.
        """
        avail_versions = self.load_remote_json(self.__class__.VERSION_URL)
        assert avail_versions[
            "format"] == "1.0", "versions.json format has changed: %s" % avail_versions[
                "format"]
        if version == "latest":
            version = avail_versions["versions"][-1]["build_version"]
            self.logger.info("Asking for latest version, ie. '%s'" % version)
        self.logger.info(
            "Find update path to bring data from version '%s' up-to version '%s'"
            % (backend_version, version))
        file_url = urljoin(self.base_url, "%s.json" % version)
        build_meta = self.load_remote_json(file_url)
        if not build_meta:
            raise Exception("Can't get remote build information about version '%s' (url was '%s')" %
                            (version, file_url))
        self.check_compat(build_meta)

        if build_meta["target_version"] == backend_version:
            self.logger.info("Backend is up-to-date, version '%s'" %
                             backend_version)
            return []

        # from older to older, each required_version being compatible with the next target_version
        # except when version is a full update, ie. no previous version required (end of the path)
        path = [build_meta]

        # latest points to a new version, 2 options there:
        # - update is a full snapshot: nothing to download if type=s3, one archive to download if type=fs, we then need to trigger a restore
        # - update is an incremental, we need to check if the incremental is compatible with current version
        if build_meta["type"] == "incremental":
            # require_version contains the compatible version for which we can apply the diff
            # let's compare...
            if backend_version == build_meta["require_version"]:
                self.logger.info("Diff update '%s' requires version '%s', which is compatible with current backend version, download update" %
                                 (build_meta["build_version"], build_meta["require_version"]))
            else:
                self.logger.info("Diff '%s' update requires version '%s'" %
                                 (build_meta["build_version"],
                                  build_meta["require_version"]))
                self.logger.info("Now looking for a compatible version")
                # by default we'll check directly the required version
                compatibles = [
                    v for v in avail_versions["versions"]
                    if v["target_version"] == version.split(".")[0]
                ]
                self.logger.info(
                    "Compatible versions from which we can apply this update: %s"
                    % compatibles)
                best_version = self.choose_best_version(compatibles)
                self.logger.info("Best version found: '%s'" % best_version)
                # keep this version as part of the update path
                # fill the path from older to newer (no extend or append)
                path = self.find_update_path(best_version["build_version"],
                                             backend_version) + path
        else:
            # full, just keep it as-is, it's a full (and it's already part of path during init, see above
            pass

        return path

    def create_todump_list(self, force=False, version="latest", url=None):
        assert self.__class__.VERSION_URL, "VERSION_URL class attribute is not set"
        self.logger.info("Dumping version '%s'" % version)
        file_url = url or urljoin(self.base_url, "%s.json" % version)
        # if "latest", we compare current json file we have (because we don't know what's behind latest)
        # otherwise json file should match version explicitely in current folder.
        version = version == "latest" and self.current_release or version
        try:
            current_localfile = os.path.join(self.current_data_folder,
                                             "%s.json" % version)
            # check it actually exists (if data folder was deleted by src_dump still refers to
            # this folder, this file won't exist)
            if not os.path.exists(current_localfile):
                self.logger.error("Local file '%s' doesn't exist" %
                                  current_localfile)
                raise FileNotFoundError
            dump_status = self.src_doc.get("download", {}).get("status")
            if dump_status != "success":
                self.logger.error(
                    "Found dump information but status is '%s', will ignore current dump"
                    % dump_status)
                raise TypeError
        except (TypeError, FileNotFoundError):
            # current data folder doesn't even exist
            current_localfile = None
        self.logger.info("Local file: %s" % current_localfile)
        self.logger.info("Remote url : %s" % file_url)
        self.logger.info("Force: %s" % force)
        if force or current_localfile is None or self.remote_is_better(
                file_url, current_localfile):
            # manually get the diff meta file (ie. not using download() because we don't know the version yet,
            # it's in the diff meta
            build_meta = self.load_remote_json(file_url)
            self.check_compat(build_meta)
            if not build_meta:
                raise Exception("Can't get remote build information about version '%s' (url was '%s')" %
                                (version, file_url))
            if build_meta["type"] == "incremental":
                self.release = build_meta["build_version"]
                # ok, now we can use download()
                # we will download it again during the normal process so we can then compare
                # when we have new data release
                new_localfile = os.path.join(self.new_data_folder,
                                             "%s.json" % self.release)
                self.to_dump.append({
                    "remote": file_url,
                    "local": new_localfile
                })
                # get base url (used later to get diff files)
                metadata_url = build_meta["metadata"]["url"]
                base_url = os.path.dirname(
                    metadata_url
                ) + "/"  # "/" or urljoin will remove previous fragment...
                new_localfile = os.path.join(self.new_data_folder,
                                             os.path.basename(metadata_url))
                self.download(metadata_url, new_localfile)
                metadata = json.load(open(new_localfile))
                remote_files = metadata["diff"]["files"]
                if metadata["diff"]["mapping_file"]:
                    remote_files.append(metadata["diff"]["mapping_file"])
                for md5_fname in remote_files:
                    fname = md5_fname["name"]
                    p = urlparse(fname)
                    if not p.scheme:
                        # this is a relative path
                        furl = urljoin(base_url, fname)
                    else:
                        # this is a true URL
                        furl = fname
                    new_localfile = os.path.join(self.new_data_folder,
                                                 os.path.basename(fname))
                    self.to_dump.append({
                        "remote": furl,
                        "local": new_localfile
                    })

            else:
                # it's a full snapshot release, it always can be applied
                self.release = build_meta["build_version"]
                new_localfile = os.path.join(self.new_data_folder,
                                             "%s.json" % self.release)
                self.to_dump.append({
                    "remote": file_url,
                    "local": new_localfile
                })
                # if repo type is fs, we assume metadata contains url to archive
                repo_name = list(
                    build_meta["metadata"]["repository"].keys())[0]
                if build_meta["metadata"]["repository"][repo_name][
                        "type"] == "fs":
                    archive_url = build_meta["metadata"]["archive_url"]
                    archive = os.path.basename(archive_url)
                    new_localfile = os.path.join(self.new_data_folder, archive)
                    self.to_dump.append({
                        "remote": archive_url,
                        "local": new_localfile
                    })

            # unset this one, as it may not be pickelable (next step is "download", which
            # uses different processes and need workers to be pickled)
            self._target_backend = None
            return self.release

    def post_dump(self, *args, **kwargs):
        if not self.release:
            # wasn't set before, means no need to post-process (ie. up-to-date, already done)
            return
        build_meta = json.load(
            open(os.path.join(self.new_data_folder, "%s.json" % self.release)))
        if build_meta["type"] == "incremental":
            self.logger.info("Checking md5sum for files in '%s'" %
                             self.new_data_folder)
            metadata = json.load(
                open(os.path.join(self.new_data_folder, "metadata.json")))
            for md5_fname in metadata["diff"]["files"]:
                spec_md5 = md5_fname["md5sum"]
                fname = md5_fname["name"]
                compute_md5 = md5sum(os.path.join(self.new_data_folder, fname))
                if compute_md5 != spec_md5:
                    self.logger.error(
                        "md5 check failed for file '%s', it may be corrupted" %
                        fname)
                    e = DumperException("Bad md5sum for file '%s'" % fname)
                    self.register_status("failed", download={"err": repr(e)})
                    raise e
                else:
                    self.logger.debug("md5 check success for file '%s'" %
                                      fname)
        elif build_meta["type"] == "full":
            # if type=fs, check if archive must be uncompressed
            repo_name = list(build_meta["metadata"]["repository"].keys())[0]
            if build_meta["metadata"]["repository"][repo_name]["type"] == "fs":
                uncompressall(self.new_data_folder)

    @asyncio.coroutine
    def info(self, version='latest'):
        """Display version information (release note, etc...) for given version"""
        file_url = urljoin(self.base_url, "%s.json" % version)
        result = {}
        build_meta = self.load_remote_json(file_url)
        if not build_meta:
            raise DumperException("Can't find version '%s'" % version)
        result["info"] = build_meta
        if build_meta.get("changes"):
            result["release_note"] = {}
            for filtyp in build_meta["changes"]:
                relnote_url = build_meta["changes"][filtyp]["url"]
                res = self.client.get(relnote_url)
                if res.status_code == 200:
                    if filtyp == "json":
                        result["release_note"][filtyp] = res.json()
                    else:
                        result["release_note"][filtyp] = res.text
                else:
                    raise DumperException(
                        "Error while downloading release note '%s (%s)': %s" %
                        (version, res, res.text))
        return result

    @asyncio.coroutine
    def versions(self):
        """Display all available versions"""
        avail_versions = self.load_remote_json(self.__class__.VERSION_URL)
        if not avail_versions:
            raise DumperException("Can't find any versions available...")
        assert avail_versions[
            "format"] == "1.0", "versions.json format has changed: %s" % avail_versions[
                "format"]
        return avail_versions["versions"]



"""
    A simple Biothings API implementation.

    * Process command line arguments to setup the API.
    * Add additional applicaion settings like handlers.

    * ``port``: the port to start the API on, **default** 8000
    * ``debug``: start the API in debug mode, **default** False
    * ``address``: the address to start the API on, **default** 0.0.0.0
    * ``autoreload``: restart the server when file changes, **default** False
    * ``conf``: choose an alternative setting, **default** config
    * ``dir``: path to app directory. **default**: current working directory

"""
import logging
import os
import sys

from tornado.options import Error, define, options

from biothings.web import BiothingsAPI

__USE_WSGI__ = False

define("port", default=8000, help="run on the given port")
define("debug", default=False, help="debug settings like logging preferences")
define("address", default=None, help="host address to listen to, default to all interfaces")
define("autoreload", default=False, help="auto reload the web server when file change detected")
define("conf", default='config', help="specify a config module name to import")
define("dir", default=os.getcwd(), help="path to app directory that includes config.py")

try:
    options.parse_command_line()
    _path = os.path.abspath(options.dir)
    if _path not in sys.path:
        sys.path.append(_path)
    del _path
except Error:
    pass  # TODO
else:
    pass


def main(app_handlers=None, app_settings=None, use_curl=False):
    """ Start a Biothings API Server

        :param app_handlers: additional web handlers to add to the app
        :param app_settings: `Tornado application settings dictionary
        <http://www.tornadoweb.org/en/stable/web.html#tornado.web.Application.settings>`_
        :param use_curl: Overide the default simple_httpclient with curl_httpclient
        <https://www.tornadoweb.org/en/stable/httpclient.html>
    """
    app_handlers = app_handlers or []
    app_settings = app_settings or {}
    api = BiothingsAPI(options.conf)

    if app_settings:
        api.settings.update(app_settings)
    if app_handlers:
        api.handlers = app_handlers
    if use_curl:
        api.use_curl()

    api.host = options.address
    api.update(debug=options.debug)
    api.update(autoreload=options.autoreload)
    api.start(options.port)


if __name__ == '__main__':
    main()



import time
import copy
from datetime import datetime
import json
from pprint import pformat
import asyncio
from functools import partial

from biothings.utils.hub_db import get_src_build
import biothings.utils.aws as aws
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager, BaseStatusRegisterer
from biothings.utils.es import ESIndexer, IndexerException as ESIndexerException
from biothings import config as btconfig
from config import logger as logging
from biothings.utils.hub import template_out
from biothings.hub import SNAPSHOOTER_CATEGORY, SNAPSHOTMANAGER_CATEGORY


class Snapshooter(BaseStatusRegisterer):
    def __init__(self, job_manager, index_manager, envconf, log_folder, *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.index_manager = index_manager
        self.job_manager = job_manager
        self.envconf = envconf
        self.log_folder = log_folder
        self.ti = time.time()
        self.setup()
        # extract some "special" values depending on the config type, etc...
        if self.envconf.get("repository", {}).get("type") == "fs":
            self.es_backups_folder = self.envconf["repository"][
                "es_backups_folder"]

    @property
    def collection(self):
        return get_src_build()

    def setup(self):
        self.setup_log()

    def setup_log(self):
        self.logger, self.logfile = get_logger(SNAPSHOOTER_CATEGORY,
                                               self.log_folder)

    def get_predicates(self):
        return []

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "category": SNAPSHOOTER_CATEGORY,
            "source": "",
            "step": "",
            "description": ""
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    def load_build(self, index_name):
        return self.load_doc(index_name, "index")

    def template_out_conf(self, build_doc):
        """
        Template out for special value using build_doc
        Templated values can look like:
            "base_path" : "onefolder/%(_meta.build_version)s"
        where "_meta.build_version" value is taken from build_doc dictionary
        (dot field notation).  In other words, such repo config are dynamic
        and potentially change for each index/snapshot created.
        """
        try:
            strconf = template_out(json.dumps(self.envconf), build_doc)
            return json.loads(strconf)
        except Exception as e:
            self.logger.exception("Coudn't template out configuration: %s" % e)
            raise

    def reset_repository_info(self, bdoc, snapshot_name):
        if bdoc.get("snapshot", {}).get(snapshot_name):
            if bdoc["snapshot"][snapshot_name].pop("repository", None):
                self.collection.save(bdoc)

    def register_status(self,
                        bdoc,
                        status,
                        transient=False,
                        init=False,
                        **extra):
        super().register_status(bdoc,
                                "snapshot",
                                status,
                                transient=transient,
                                init=init,
                                **extra)

    def get_es_idxr(self, envconf, index=None):
        if envconf["indexer"].get("env"):
            # we should take indexer params from INDEX_CONFIG, ie. index_manager
            idxklass = self.index_manager.find_indexer(index)
            idxkwargs = self.index_manager[envconf["indexer"]["env"]]
        else:
            idxklass = self.index_manager.DEFAULT_INDEXER
            idxkwargs = envconf["indexer"]["host"]["args"]
        idxr = idxklass(**idxkwargs)
        es_idxr = ESIndexer(index=index,
                            doc_type=idxr.doc_type,
                            es_host=idxr.host,
                            check_index=index is not None)

        return es_idxr

    def pre_snapshot(self, envconf, repo_conf, index_meta):
        pass

    def snapshot(self,
                 index,
                 snapshot=None,
                 steps=["pre", "snapshot", "post"]):
        bdoc = self.load_build(index)
        self.envconf = self.template_out_conf(bdoc)
        # check what to do
        if type(steps) == str:
            steps = [steps]
        snapshot_name = snapshot or index
        es_idxr = self.get_es_idxr(self.envconf, index)
        # create repo if needed
        index_meta = es_idxr.get_mapping_meta()["_meta"]  # read from index
        repo_name, repo_conf = self.create_repository(self.envconf, bdoc,
                                                      index_meta)
        monitor_delay = self.envconf["monitor_delay"]
        # will hold the overall result
        fut = asyncio.Future()

        def get_status():
            try:
                res = es_idxr.get_snapshot_status(repo_name, snapshot_name)
                assert "snapshots" in res, "Can't find snapshot '%s' in repository '%s'" % (
                    snapshot_name, repo_name)
                # assuming only one index in the snapshot, so only check first elem
                info = res["snapshots"][0]
                assert info.get(
                    "state"
                ), "Can't find state in snapshot '%s'" % snapshot_name
                return info
            except Exception:
                # something went wrong, report as failure
                return "FAILED"

        @asyncio.coroutine
        def do(index):
            try:
                global_state = None
                got_error = None

                def snapshot_launched(f):
                    try:
                        self.logger.info("Snapshot launched: %s" % f.result())
                    except Exception as e:
                        self.logger.error("Error while lauching snapshot: %s" %
                                          e)
                        nonlocal got_error
                        got_error = e

                def done(f, step):
                    try:
                        res = f.result()
                        self.register_status(bdoc,
                                             "success",
                                             job={
                                                 "step": "%s-snapshot" % step,
                                                 "result": res
                                             },
                                             snapshot={
                                                 snapshot_name: {
                                                     "conf": self.envconf,
                                                     step: res
                                                 }
                                             })
                        self.logger.info("%s-snapshot done: %s" % (step, res))
                    except Exception as e:
                        nonlocal got_error
                        got_error = e
                        self.register_status(bdoc,
                                             "failed",
                                             job={
                                                 "step": "%s-snapshot" % step,
                                                 "err": str(e)
                                             },
                                             snapshot={
                                                 snapshot_name: {
                                                     "conf": self.envconf,
                                                     step: None
                                                 }
                                             })
                        self.logger.exception(
                            "Error while running pre-snapshot: %s" % e)

                pinfo = self.get_pinfo()
                pinfo["source"] = index

                # we only allow one repo conf per snapshot name
                self.reset_repository_info(bdoc, snapshot_name)

                if "pre" in steps:
                    self.register_status(bdoc,
                                         "pre-snapshotting",
                                         transient=True,
                                         init=True,
                                         job={"step": "pre-snapshot"},
                                         snapshot={snapshot_name: {}})
                    pinfo["step"] = "pre-snapshot"
                    pinfo.pop("description", None)
                    job = yield from self.job_manager.defer_to_thread(
                        pinfo,
                        partial(self.pre_snapshot, self.envconf, repo_conf,
                                index_meta))
                    job.add_done_callback(partial(done, step="pre"))
                    yield from job
                    if got_error:
                        fut.set_exception(got_error)
                        return

                if "snapshot" in steps:
                    self.register_status(bdoc,
                                         "snapshotting",
                                         transient=True,
                                         init=True,
                                         job={"step": "snapshot"},
                                         snapshot={snapshot_name: {}})
                    pinfo["step"] = "snapshot"
                    pinfo["description"] = es_idxr.es_host
                    self.logger.info(
                        "Creating snapshot for index '%s' on host '%s', repository '%s'"
                        % (index, es_idxr.es_host, repo_name))
                    job = yield from self.job_manager.defer_to_thread(
                        pinfo,
                        partial(es_idxr.snapshot, repo_name, snapshot_name))
                    job.add_done_callback(snapshot_launched)
                    yield from job

                    # launched, so now monitor completion
                    while True:
                        info = get_status()
                        state = info["state"]
                        failed_shards = info["shards_stats"]["failed"]
                        if state in ["INIT", "IN_PROGRESS", "STARTED"]:
                            yield from asyncio.sleep(monitor_delay)
                        else:
                            if state == "SUCCESS" and failed_shards == 0:
                                global_state = state.lower()
                                self.logger.info("Snapshot '%s' successfully created (host: '%s', repository: '%s')" %
                                                 (snapshot_name, es_idxr.es_host, repo_name), extra={"notify": True})
                            else:
                                if state == "SUCCESS":
                                    e = IndexerException(
                                        "Snapshot '%s' partially failed: state is %s but %s shards failed"
                                        %
                                        (snapshot_name, state, failed_shards))
                                else:
                                    e = IndexerException(
                                        "Snapshot '%s' failed: state is %s" %
                                        (snapshot_name, state))
                                self.logger.error("Failed creating snapshot '%s' (host: %s, repository: %s), state: %s" %
                                                  (snapshot_name, es_idxr.es_host, repo_name, state), extra={"notify": True})
                                got_error = e
                            break

                    if got_error:
                        self.register_status(bdoc,
                                             "failed",
                                             job={
                                                 "step": "snapshot",
                                                 "err": str(got_error)
                                             },
                                             snapshot={
                                                 snapshot_name: {
                                                     "conf": self.envconf,
                                                     "repository": None,
                                                 }
                                             })
                        fut.set_exception(got_error)
                        return

                    else:
                        self.register_status(bdoc,
                                             "success",
                                             job={
                                                 "step": "snapshot",
                                                 "status": global_state,
                                             },
                                             snapshot={
                                                 snapshot_name: {
                                                     "conf": self.envconf,
                                                     "repository": {
                                                         repo_name: repo_conf
                                                     }
                                                 }
                                             })

                if "post" in steps:
                    self.register_status(bdoc,
                                         "post-snapshotting",
                                         transient=True,
                                         init=True,
                                         job={"step": "post-snapshot"},
                                         snapshot={snapshot_name: {}})
                    pinfo["step"] = "post-snapshot"
                    pinfo.pop("description", None)
                    job = yield from self.job_manager.defer_to_thread(
                        pinfo,
                        partial(self.post_snapshot, self.envconf, repo_conf,
                                index_meta))
                    job.add_done_callback(partial(done, step="post"))
                    yield from job
                    if got_error:
                        fut.set_exception(got_error)
                        return

                # if  we get there all is fine
                fut.set_result(global_state)

            except Exception as e:
                self.logger.exception(e)
                fut.set_exception(e)

        asyncio.ensure_future(do(index))
        return fut

    def post_snapshot(self, envconf, repo_conf, index_meta):
        pass

    def get_repository_config(self, repo_conf, build_doc, index_meta):
        repo_name = repo_conf["name"]
        repo_type = repo_conf["type"]
        repo_settings = {}
        for setting in repo_conf["settings"]:
            repo_settings[setting] = repo_conf["settings"][setting]
        return repo_name, {"type": repo_type, "settings": repo_settings}

    def create_repository(self, envconf, build_doc, index_meta):
        aws_key = envconf.get("cloud", {}).get("access_key")
        aws_secret = envconf.get("cloud", {}).get("secret_key")
        repo_name, repo_conf = self.get_repository_config(
            envconf["repository"], build_doc, index_meta)
        self.logger.info("Repository config: %s" % repo_conf)
        settings = repo_conf["settings"]
        repo_type = repo_conf["type"]
        es_idxr = self.get_es_idxr(envconf)
        try:
            es_idxr.get_repository(repo_name)
        except ESIndexerException:
            # need to create that repo
            if repo_type == "s3":
                acl = repo_conf.get("acl",
                                    None)  # let aws.create_bucket default it
                # first make sure bucket exists
                aws.create_bucket(name=settings["bucket"],
                                  region=settings["region"],
                                  aws_key=aws_key,
                                  aws_secret=aws_secret,
                                  acl=acl,
                                  ignore_already_exists=True)
            settings = {"type": repo_type, "settings": settings}
            self.logger.info("Create repository named '%s': %s" %
                             (repo_name, pformat(settings)))
            es_idxr.create_repository(repo_name, settings=settings)

        return repo_name, repo_conf


class SnapshotManager(BaseManager):

    DEFAULT_SNAPSHOOTER_CLASS = Snapshooter

    def __init__(self, index_manager, poll_schedule=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index_manager = index_manager
        self.t0 = time.time()
        self.log_folder = btconfig.LOG_FOLDER
        self.timestamp = datetime.now()
        self.snapshot_config = {}
        self.poll_schedule = poll_schedule
        self.es_backups_folder = getattr(btconfig, "ES_BACKUPS_FOLDER", None)
        self.setup()

    def clean_stale_status(self):
        src_build = get_src_build()
        for build in src_build.find():
            for job in build.get("jobs", []):
                if job.get("status", "").endswith("snapshotting"):
                    logging.warning(
                        "Found stale build '%s', marking snapshot status as 'canceled'"
                        % build["_id"])
                    job["status"] = "canceled"
            src_build.replace_one({"_id": build["_id"]}, build)

    def setup(self):
        self.setup_log()

    def setup_log(self):
        self.logger, self.logfile = get_logger(SNAPSHOTMANAGER_CATEGORY,
                                               self.log_folder)

    def poll(self, state, func):
        super().poll(state, func, col=get_src_build())

    def __getitem__(self, env):
        """
        Return an instance of a snapshooter for snapshot environment named "env"
        """
        d = self.register[env]
        envconf = self.snapshot_config["env"][env]
        return d["class"](index_manager=d["index_manager"],
                          job_manager=d["job_manager"],
                          envconf=envconf,
                          log_folder=self.log_folder)

    def configure(self, snapshot_confdict):
        """
        Configure manager with snapshot config dict. See SNAPSHOT_CONFIG in config_hub.py
        for the format.
        """
        self.snapshot_config = copy.deepcopy(snapshot_confdict)
        for env, envconf in self.snapshot_config.get("env", {}).items():
            try:
                if envconf.get("cloud"):
                    assert envconf["cloud"]["type"] == "aws", \
                        "Only Amazon AWS cloud is supported at the moment"
                self.register[env] = {
                    "class": self.DEFAULT_SNAPSHOOTER_CLASS,
                    "conf": envconf,
                    "job_manager": self.job_manager,
                    "index_manager": self.index_manager
                }
            except Exception as e:
                self.logger.exception(
                    "Couldn't setup snapshot environment '%s' because: %s" %
                    (env, e))

    def snapshot(self,
                 snapshot_env,
                 index,
                 snapshot=None,
                 steps=["pre", "snapshot", "post"]):
        """
        Create a snapshot named "snapshot" (or, by default, same name as the index) from "index"
        according to environment definition (repository, etc...) "snapshot_env".

        """
        if snapshot_env not in self.snapshot_config.get("env", {}):
            raise ValueError("Unknonw snapshot environment '%s'" %
                             snapshot_env)
        snapshooter = self[snapshot_env]
        return snapshooter.snapshot(index, snapshot=snapshot, steps=steps)

    def snapshot_info(self, env=None, remote=False):
        res = copy.deepcopy(self.snapshot_config)
        for kenv in self.snapshot_config["env"]:
            if env and env != kenv:
                continue
            if remote:
                raise NotImplementedError()
                # TODO: could list snapshot from ES
                #try:
                #    cl = Elasticsearch(res["env"][kenv]["host"],timeout=1,max_retries=0)
                #    indices = [{"index":k,
                #        "doc_type":list(v["mappings"].keys())[0],
                #        "aliases":list(v["aliases"].keys())}
                #        for k,v in cl.indices.get("*").items()]
                #    # for now, we just consider
                #    if type(res["env"][kenv]["index"]) == dict:
                #        # we don't where to put those indices because we don't
                #        # have that information, so we just put those in a default category
                #        # TODO: put that info in metadata ?
                #        res["env"][kenv]["index"].setdefault(None,[]).extend(indices)
                #    else:
                #        assert type(res["env"][kenv]["index"]) == list
                #        res["env"][kenv]["index"].extend(indices)
                #except Exception as e:
                #    self.logger.warning("Can't load remote indices: %s" % e)
                #    continue
        return res



import tornado.web
import json.decoder

from .base import GenericHandler
from biothings.utils.hub import CommandError, NoSuchCommand, CommandNotAllowed


class ShellHandler(GenericHandler):
    def initialize(self, shell, shellog, **kwargs):
        super().initialize(shell, **kwargs)
        self.shellog = shellog

    def put(self):
        try:
            bodyargs = tornado.escape.json_decode(self.request.body or '{}')
            cmd = bodyargs["cmd"]
        except KeyError:
            raise tornado.web.HTTPError(
                400, reason="Bad Request oula (Missing argument cmd)")
        except json.decoder.JSONDecodeError:
            raise tornado.web.HTTPError(400, reason="Invalid JSON payload")

        try:
            outs = self.shell.eval(cmd,
                                   secure=True)  # only pre-defined command
            for out in outs:
                if out != "":
                    self.shellog.output(out)
        except CommandError as e:
            raise tornado.web.HTTPError(400, reason="Error: %s" % e)
        except NoSuchCommand as e:
            raise tornado.web.HTTPError(404, reason="No such command: %s" % e)
        except CommandNotAllowed as e:
            raise tornado.web.HTTPError(403,
                                        reason="Command not allowed: %s" % e)



import asyncio
import crypt
import os
import sys
import types
import time
import copy
import logging
import glob
from collections import OrderedDict
from functools import partial
from pprint import pformat

import asyncssh
import aiocron

from biothings import config
from biothings.utils.loggers import get_logger, WSLogHandler, WSShellHandler, ShellLogger
from biothings.utils.hub import (HubShell, get_hub_reloader, CommandDefinition, pending,
                                 AlreadyRunningException, CommandError)
from biothings.utils.jsondiff import make as jsondiff
from biothings.utils.version import check_new_version, get_version
from biothings.utils.common import get_class_from_classpath

# Keys used as category in pinfo (description of jobs submitted to JobManager)
# Those are used in different places
DUMPER_CATEGORY = "dumper"
UPLOADER_CATEGORY = "uploader"
BUILDER_CATEGORY = "builder"
INDEXER_CATEGORY = "indexer"
INDEXMANAGER_CATEGORY = "indexmanager"
RELEASEMANAGER_CATEGORY = "releasemanager"
RELEASER_CATEGORY = "releaser"
SNAPSHOTMANAGER_CATEGORY = "snapshotmanager"
SNAPSHOOTER_CATEGORY = "snapshooter"
DIFFER_CATEGORY = "differ"
DIFFMANAGER_CATEGORY = "diffmanager"
SYNCER_CATEGORY = "syncer"
INSPECTOR_CATEGORY = "inspector"

# HUB_REFRESH_COMMANDS = hasattr(
#     config, "HUB_REFRESH_COMMANDS"
# ) and config.HUB_REFRESH_COMMANDS or "* * * * * *"  # every sec
HUB_REFRESH_COMMANDS = getattr(
    config, "HUB_REFRESH_COMMANDS", "* * * * * *"  # every sec
)

# Check for new code update from app and biothings Git repo
HUB_CHECK_UPGRADE = getattr(
    config, "HUB_CHECK_UPGRADE", "0 * * * *"  # every hour
)


class JobRenderer(object):
    def __init__(self):
        self.rendered = {
            types.FunctionType: self.render_func,
            types.MethodType: self.render_method,
            partial: self.render_partial,
            types.LambdaType: self.render_lambda,
        }

    def render(self, job):
        r = self.rendered.get(type(job._callback))
        rstr = r(job._callback)
        delta = job._when - job._loop.time()
        days = None
        if delta > 86400:
            days = int(delta / 86400)
            delta = delta - 86400
        strdelta = time.strftime("%Hh:%Mm:%Ss", time.gmtime(int(delta)))
        if days:
            strdelta = "%d day(s) %s" % (days, strdelta)
        return "%s {run in %s}" % (rstr, strdelta)

    def render_partial(self, p):
        # class.method(args)
        return self.rendered[type(p.func)](p.func) + "%s" % str(p.args)

    def render_cron(self, c):
        # func type associated to cron can vary
        return self.rendered[type(c.func)](c.func) + " [%s]" % c.spec

    def render_func(self, f):
        return f.__name__

    def render_method(self, m):
        # what is self ? cron ?
        # if type(m.__self__) == aiocron.Cron:   # TODO: delete if confirmed
        if isinstance(m.__self__, aiocron.Cron):
            return self.render_cron(m.__self__)
        else:
            return "%s.%s" % (m.__self__.__class__.__name__, m.__name__)

    def render_lambda(self, l):
        return l.__name__


renderer = JobRenderer()


def status(managers):
    """
    Return a global hub status (number or sources, documents, etc...)
    according to available managers
    """
    total_srcs = None
    total_docs = None
    total_confs = None
    total_builds = None
    total_apis = None
    total_running_apis = None
    if managers.get("source_manager"):
        try:
            srcm = managers["source_manager"]
            srcs = srcm.get_sources()
            total_srcs = len(srcs)
            total_docs = sum([s["upload"]["sources"][subs].get("count", 0) or 0
                              for s in srcs
                              for subs in s.get("upload", {}).get("sources", {})
                              if s.get("upload")])
        except Exception:
            logging.exception("Can't get stats for sources:")

    try:
        bm = managers["build_manager"]
        total_confs = len(bm.build_config_info())
    except Exception:
        logging.exception("Can't get total number of build configurations:")
    try:
        total_builds = len(bm.build_info())
    except Exception:
        logging.exception("Can't get total number of builds:")

    try:
        am = managers["api_manager"]
        apis = am.get_apis()
        total_apis = len(apis)
        total_running_apis = len(
            [a for a in apis if a.get("status") == "running"])
    except Exception:
        logging.exception("Can't get stats for APIs:")

    return {
        "source": {
            "total": total_srcs,
            "documents": total_docs
        },
        "build": {
            "total": total_builds
        },
        "build_conf": {
            "total": total_confs
        },
        "api": {
            "total": total_apis,
            "running": total_running_apis
        },
    }


def get_schedule(loop):
    """try to render job in a human-readable way..."""
    out = []
    for sch in loop._scheduled:
        if type(sch) != asyncio.events.TimerHandle:
            continue
        if sch._cancelled:
            continue
        try:
            info = renderer.render(sch)
            out.append(info)
        except Exception:
            import traceback
            traceback.print_exc()
            out.append(sch)

    return "\n".join(out)


@asyncio.coroutine
def start_ssh_server(loop,
                     name,
                     passwords,
                     keys=['bin/ssh_host_key'],
                     shell=None,
                     host='',
                     port=8022):
    for key in keys:
        assert os.path.exists(
            key
        ), "Missing key '%s' (use: 'ssh-keygen -f %s' to generate it" % (key,
                                                                         key)
    HubSSHServer.PASSWORDS = passwords
    HubSSHServer.NAME = name
    HubSSHServer.SHELL = shell
    aiocron.crontab(HUB_REFRESH_COMMANDS,
                    func=shell.__class__.refresh_commands,
                    start=True,
                    loop=loop)
    yield from asyncssh.create_server(HubSSHServer,
                                      host,
                                      port,
                                      loop=loop,
                                      server_host_keys=keys)


class HubCommands(OrderedDict):
    def __setitem__(self, k, v):
        if k in self:
            raise ValueError("Command '%s' already defined" % k)
        super().__setitem__(k, v)


class HubServer(object):

    DEFAULT_FEATURES = [
        "config", "job", "dump", "upload", "dataplugin", "source", "build",
        "diff", "index", "snapshot", "release", "inspect", "sync", "api",
        "terminal", "reloader", "dataupload", "ws", "readonly", "upgrade",
        "autohub","hooks",
    ]
    DEFAULT_MANAGERS_ARGS = {"upload": {"poll_schedule": "* * * * * */10"}}
    DEFAULT_RELOADER_CONFIG = {
        "folders": None,  # will use default one
        "managers": ["source_manager", "assistant_manager"],
        "reload_func": None
    }  # will use default one
    DEFAULT_DATAUPLOAD_CONFIG = {
        "upload_root": getattr(config, "DATA_UPLOAD_FOLDER", None)
    }
    DEFAULT_WEBSOCKET_CONFIG = {}
    DEFAULT_API_CONFIG = {}
    DEFAULT_AUTOHUB_CONFIG = {
        "version_urls": getattr(config, "VERSION_URLS", []),
        "indexer_factory": getattr(config, "AUTOHUB_INDEXER_FACTORY", None),
        "es_host": getattr(config, "AUTOHUB_ES_HOST", None),
    }

    def __init__(self,
                 source_list,
                 features=None,
                 name="BioThings Hub",
                 managers_custom_args={},
                 api_config=None,
                 reloader_config=None,
                 dataupload_config=None,
                 websocket_config=None,
                 autohub_config=None):
        """
        Helper to setup and instantiate common managers usually used in a hub
        (eg. dumper manager, uploader manager, etc...)
        "source_list" is either:
            - a list of string corresponding to paths to datasources modules
            - a package containing sub-folders with datasources modules
        Specific managers can be retrieved adjusting "features" parameter, where
        each feature corresponds to one or more managers. Parameter defaults to
        all possible available. Managers are configured/init in the same order as the list,
        so if a manager (eg. job_manager) is required by all others, it must be the first
        in the list.
        "managers_custom_args" is an optional dict used to pass specific arguments while
        init managers:
            managers_custom_args={"upload" : {"poll_schedule" : "*/5 * * * *"}}
        will set poll schedule to check upload every 5min (instead of default 10s)
        "reloader_config", "dataupload_config", "autohub_config" and "websocket_config"
        can be used to customize reloader, dataupload and websocket. If None, default config
        is used. If explicitely False, feature is deactivated.
        """
        self.name = name
        self.source_list = source_list
        self.logger, self.logfile = get_logger("hub")
        self._passed_features = features
        self._passed_managers_custom_args = managers_custom_args
        self.features = self.clean_features(features or self.DEFAULT_FEATURES)
        self.managers_custom_args = managers_custom_args
        self.reloader_config = reloader_config or self.DEFAULT_RELOADER_CONFIG
        self.dataupload_config = dataupload_config or self.DEFAULT_DATAUPLOAD_CONFIG
        self.websocket_config = websocket_config or self.DEFAULT_WEBSOCKET_CONFIG
        self.autohub_config = autohub_config or self.DEFAULT_AUTOHUB_CONFIG
        self.ws_listeners = [
        ]  # collect listeners that should be connected (push data through) to websocket
        self.api_config = api_config or self.DEFAULT_API_CONFIG
        # set during configure()
        self.managers = None
        self.api_endpoints = {}
        self.readonly_api_endpoints = None
        self.shell = None
        self.commands = None  # default "public" commands
        self.extra_commands = None  # "hidden" commands, but still useful for advanced usage
        self.hook_files = None  # user-defined commands as hook files
        self.routes = []
        self.readonly_routes = []
        self.ws_urls = []  # only one set, shared between r/w and r/o hub api server
        # flag "do we need to configure?"
        self.configured = False

    def clean_features(self, features):
        """
        Sanitize (ie. remove duplicates) features
        """
        # we can't just use "set()" because we need to preserve order
        ordered = OrderedDict()
        for feat in features:
            if feat not in ordered:
                ordered[feat] = None
        return list(ordered.keys())

    def before_configure(self):
        """
        Hook triggered before configure(),
        used eg. to adjust features list
        """
        pass

    def configure_readonly_api_endpoints(self):
        """
        Assuming read-write API endpoints have previously been defined (self.api_endpoints set)
        extract commands and their endpoint definitions only when method is GET. That is, for any
        given API definition honoring REST principle for HTTP verbs, generate endpoints only for
        which actions are read-only actions.
        """
        assert self.api_endpoints, "Can't derive a read-only API is no read-write endpoints are defined"
        self.readonly_api_endpoints = {}
        for cmd, api_endpoints in self.api_endpoints.items():
            if not isinstance(api_endpoints, list):
                api_endpoints = [api_endpoints]
            for endpoint in api_endpoints:
                if endpoint["method"].lower() != "get":
                    self.logger.debug("Skipping %s: %s for read-only API" % (cmd, endpoint))
                    continue
                else:
                    self.readonly_api_endpoints.setdefault(cmd, []).append(endpoint)

    def configure(self):
        self.before_configure()
        self.remaining_features = copy.deepcopy(
            self.features)  # keep track of what's been configured
        self.configure_ioloop()
        self.configure_managers()
        # setup the shell
        self.shell = HubShell(self.managers["job_manager"])
        self.shell.register_managers(self.managers)
        self.shell.server = self  # propagate server instance in shell
        # so it's accessible from the console if needed
        self.configure_remaining_features()
        self.configure_commands()
        self.configure_extra_commands()
        self.shell.set_commands(self.commands, self.extra_commands)
        self.ingest_hooks()
        # setapi
        if self.api_config is not False:
            self.configure_api_endpoints(
            )  # after shell setup as it adds some default commands
            # we want to expose throught the api
            from biothings.hub.api import generate_api_routes
            from biothings.hub.api.handlers.base import RootHandler

            # First deal with read-only API
            if "readonly" in self.features:
                self.configure_readonly_api_endpoints()
                self.readonly_routes.extend(
                    generate_api_routes(self.shell, self.readonly_api_endpoints))
                # we don't want to expose feature read-only for the API that is *not*
                # read-only. "readonly" feature means we're running another webapp for
                # a specific readonly API. UI can then query the root handler and see
                # if the API is readonly or not, and adjust the components & actions
                ro_features = copy.deepcopy(self.features)
                # terminal feature certainly not allowed in read-only server...
                if "terminal" in self.features:
                    ro_features.remove("terminal")
                # if we have readonly feature, it means another non-readonly server is running
                self.features.remove("readonly")
                hub_name = getattr(config, "HUB_NAME", "Hub") + " (read-only)"
                self.readonly_routes.append(("/", RootHandler, {
                    "features": ro_features, "hub_name": hub_name
                }))

            # Then deal with read-write API
            self.routes.extend(
                generate_api_routes(self.shell, self.api_endpoints))
            self.routes.append(("/", RootHandler, {
                "features": self.features,
            }))

        # done
        self.configured = True

    def configure_ioloop(self):
        import tornado.platform.asyncio
        tornado.platform.asyncio.AsyncIOMainLoop().install()

    def before_start(self):
        pass

    def start(self):
        if not self.configured:
            self.configure()
        self.logger.info("Starting '%s'", self.name)
        # can't use asyncio.get_event_loop() if python < 3.5.3 as it would return
        # another instance of aio loop, take it from job_manager to make sure
        # we share the same one
        loop = self.managers["job_manager"].loop
        if self.routes:
            self.logger.info("Starting Hub API server on port %s" % config.HUB_API_PORT)
            #self.logger.info(self.routes)
            import tornado.web
            # register app into current event loop
            api = tornado.web.Application(self.routes)
            self.extra_commands["api"] = api
            from biothings.hub.api import start_api
            start_api(api,
                      config.HUB_API_PORT,
                      settings=getattr(config, "TORNADO_SETTINGS", {})
                      )
            if self.readonly_routes:
                if not getattr(config, "READONLY_HUB_API_PORT", None):
                    self.logger.warning("Read-only Hub API feature is set but READONLY_HUB_API_PORT "
                                        + "isn't set in configuration")
                else:
                    self.logger.info("Starting read-only Hub API server on port %s" % config.READONLY_HUB_API_PORT)
                    #self.logger.info(self.readonly_routes)
                    ro_api = tornado.web.Application(self.readonly_routes)
                    start_api(ro_api,
                              config.READONLY_HUB_API_PORT,
                              settings=getattr(config, "TORNADO_SETTINGS", {}))
        else:
            self.logger.info("No route defined, API server won't start")
        # at this point, everything is ready/set, last call for customizations
        self.before_start()
        self.logger.info("Starting Hub SSH server on port %s" % config.HUB_SSH_PORT)
        self.ssh_server = start_ssh_server(loop,
                                           self.name,
                                           passwords=config.HUB_PASSWD,
                                           port=config.HUB_SSH_PORT,
                                           shell=self.shell)
        try:
            loop.run_until_complete(self.ssh_server)
        except (OSError, asyncssh.Error) as exc:
            sys.exit('Error starting server: ' + str(exc))
        loop.run_forever()

    def mixargs(self, feat, params=None):
        params = params or {}
        args = {}
        for p in params:
            args[p] = self.managers_custom_args.get(feat, {}).pop(
                p, None) or params[p]
        # mix remaining
        args.update(self.managers_custom_args.get(feat, {}))
        return args

    def configure_job_manager(self):
        import asyncio
        loop = asyncio.get_event_loop()
        from biothings.utils.manager import JobManager
        args = self.mixargs(
            "job", {
                "num_workers": config.HUB_MAX_WORKERS,
                "max_memory_usage": config.HUB_MAX_MEM_USAGE
            })
        job_manager = JobManager(loop, **args)
        self.managers["job_manager"] = job_manager

    def configure_dump_manager(self):
        from biothings.hub.dataload.dumper import DumperManager
        args = self.mixargs("dump")
        dmanager = DumperManager(job_manager=self.managers["job_manager"],
                                 **args)
        self.managers["dump_manager"] = dmanager

    def configure_upload_manager(self):
        from biothings.hub.dataload.uploader import UploaderManager
        args = self.mixargs("upload", {"poll_schedule": "* * * * * */10"})
        upload_manager = UploaderManager(
            job_manager=self.managers["job_manager"], **args)
        self.managers["upload_manager"] = upload_manager

    def configure_dataplugin_manager(self):
        from biothings.hub.dataplugin.manager import DataPluginManager
        dp_manager = DataPluginManager(
            job_manager=self.managers["job_manager"])
        self.managers["dataplugin_manager"] = dp_manager
        from biothings.hub.dataplugin.assistant import AssistantManager
        args = self.mixargs("dataplugin")
        assistant_manager = AssistantManager(
            data_plugin_manager=dp_manager,
            dumper_manager=self.managers["dump_manager"],
            uploader_manager=self.managers["upload_manager"],
            job_manager=self.managers["job_manager"],
            **args)
        self.managers["assistant_manager"] = assistant_manager

    def configure_build_manager(self):
        from biothings.hub.databuild.builder import BuilderManager
        args = self.mixargs("build")
        build_manager = BuilderManager(
            job_manager=self.managers["job_manager"], **args)
        build_manager.configure()
        self.managers["build_manager"] = build_manager
        build_manager.poll()

    def configure_diff_manager(self):
        from biothings.hub.databuild.differ import DifferManager, SelfContainedJsonDiffer
        args = self.mixargs("diff")
        diff_manager = DifferManager(job_manager=self.managers["job_manager"],
                                     poll_schedule="* * * * * */10",
                                     **args)
        diff_manager.configure([
            SelfContainedJsonDiffer,
        ])
        diff_manager.poll(
            "diff", lambda doc: diff_manager.diff(
                "jsondiff-selfcontained", old=None, new=doc["_id"]))
        self.managers["diff_manager"] = diff_manager

    def configure_index_manager(self):
        from biothings.hub.dataindex.indexer import IndexManager
        args = self.mixargs("index")
        index_manager = IndexManager(job_manager=self.managers["job_manager"],
                                     **args)
        index_manager.configure(config.INDEX_CONFIG)
        self.managers["index_manager"] = index_manager

    def configure_snapshot_manager(self):
        assert "index" in self.features, "'snapshot' feature requires 'index'"
        from biothings.hub.dataindex.snapshooter import SnapshotManager
        args = self.mixargs("snapshot")
        snapshot_manager = SnapshotManager(
            index_manager=self.managers["index_manager"],
            job_manager=self.managers["job_manager"],
            **args)
        snapshot_manager.configure(config.SNAPSHOT_CONFIG)
        #snapshot_manager.poll("snapshot",lambda doc: snapshot_manager.snapshot(snapshot_env=???,index=doc["_id"]))
        self.managers["snapshot_manager"] = snapshot_manager

    def configure_release_manager(self):
        assert "diff" in self.features, "'release' feature requires 'diff'"
        assert "snapshot" in self.features, "'release' feature requires 'snapshot'"
        from biothings.hub.datarelease.publisher import ReleaseManager
        args = self.mixargs("release")
        release_manager = ReleaseManager(
            diff_manager=self.managers["diff_manager"],
            snapshot_manager=self.managers["snapshot_manager"],
            job_manager=self.managers["job_manager"],
            poll_schedule="* * * * * */10",
            **args)
        release_manager.configure(config.RELEASE_CONFIG)
        release_manager.poll(
            "release_note",
            lambda doc: release_manager.create_release_note(old=None,
                                                            new=doc["_id"]))
        self.managers["release_manager"] = release_manager

    def configure_sync_manager(self):
        from biothings.hub.databuild.syncer import SyncerManager
        args = self.mixargs("sync")
        sync_manager = SyncerManager(job_manager=self.managers["job_manager"],
                                     **args)
        sync_manager.configure()
        self.managers["sync_manager"] = sync_manager

    def configure_inspect_manager(self):
        assert "upload" in self.features, "'inspect' feature requires 'upload'"
        assert "build" in self.features, "'inspect' feature requires 'build'"
        from biothings.hub.datainspect.inspector import InspectorManager
        args = self.mixargs("inspect")
        inspect_manager = InspectorManager(
            upload_manager=self.managers["upload_manager"],
            build_manager=self.managers["build_manager"],
            job_manager=self.managers["job_manager"],
            **args)
        self.managers["inspect_manager"] = inspect_manager

    def configure_api_manager(self):
        assert "index" in self.features, "'api' feature requires 'index'"
        from biothings.hub.api.manager import APIManager
        args = self.mixargs("api")
        api_manager = APIManager(**args)
        self.managers["api_manager"] = api_manager

    def configure_source_manager(self):
        if "dump" in self.features or "upload" in self.features:
            self.mixargs("source")
            from biothings.hub.dataload.source import SourceManager
            source_manager = SourceManager(
                source_list=self.source_list,
                dump_manager=self.managers["dump_manager"],
                upload_manager=self.managers["upload_manager"],
                data_plugin_manager=self.managers.get("dataplugin_manager"),
            )
            self.managers["source_manager"] = source_manager
        # init data plugin once source_manager has been set (it inits dumper and uploader
        # managers, if assistant_manager is configured/loaded before, datasources won't appear
        # in dumper/uploader managers as they were not ready yet)
        if "dataplugin" in self.features:
            self.managers["assistant_manager"].configure()
            self.managers["assistant_manager"].load()

        # now that we have the source manager setup, we can schedule and poll
        if "dump" in self.features and not getattr(
                config, "SKIP_DUMPER_SCHEDULE", False):
            self.managers["dump_manager"].schedule_all()
        if "upload" in self.features and not getattr(
                config, "SKIP_UPLOADER_POLL", False):
            self.managers["upload_manager"].poll(
                'upload', lambda doc: self.shell.launch(
                    partial(self.managers["upload_manager"].upload_src, doc[
                        "_id"])))

    def configure_autohub_feature(self):
        """
        See bt.hub.standalone.AutoHubFeature
        """
        # "autohub" feature is based on "dump","upload" and "sync" features.
        # If autohub is running on its own (standalone instance only for instance)
        # we don't list them in DEFAULT_FEATURES as we don't want them to produce
        # commands such as dump() or upload() as these are renamed for clarity
        # that said, those managers could still exist *if* autohub is mixed
        # with "standard" hub, so we don't want to override them if already configured
        if not self.managers.get("dump_manager"):
            self.configure_dump_manager()
        if not self.managers.get("upload_manager"):
            self.configure_upload_manager()
        if not self.managers.get("sync_manager"):
            self.configure_sync_manager()

        # Originally, autohub was a hub server on its own, it's now
        # converted a feature;to avoid mixins and bringing complexity in this HubServer
        # definition, we use composition pointing to an instance of that feature which
        # encapsulates that complexity
        from biothings.hub.standalone import AutoHubFeature
        # only pass required manage rs
        autohub_managers = {
            "dump_manager": self.managers["dump_manager"],
            "upload_manager": self.managers["upload_manager"],
            "sync_manager": self.managers["sync_manager"],
        }

        version_urls = self.autohub_config["version_urls"]
        indexer_factory = self.autohub_config["indexer_factory"]
        es_host = self.autohub_config["es_host"]
        factory = None
        if indexer_factory:
            assert es_host, "indexer_factory set but es_host not set (AUTOHUB_ES_HOST), can't know which ES server to use"
            try:
                factory_class = get_class_from_classpath(indexer_factory)
                factory = factory_class(version_urls, es_host)
            except (ImportError, ModuleNotFoundError) as e:
                self.logger.error("Couldn't find indexer factory class from '%s': %s" % (indexer_factory, e))
        self.autohub_feature = AutoHubFeature(autohub_managers, version_urls, factory)
        try:
            self.autohub_feature.configure()
        except Exception as e:
            self.logger.error("Could't configure feature 'autohub', will be deactivated: %s" % e)
            self.features.remove("autohub")

    def configure_hooks_feature(self):
        """
        Ingest user-defined commands into hub namespace, giving access
        to all pre-defined commands (commands, extra_commands).
        This method prepare the hooks but the ingestion is done later
        when all commands are defined
        """
        hooks_folder = getattr(config, "HOOKS_FOLDER", "./hooks")
        if not os.path.exists(hooks_folder):
            self.logger.info("Hooks folder '%s' doesn't exist, creating it" % hooks_folder)
            os.makedirs(hooks_folder)
        self.hook_files = glob.glob(os.path.join(hooks_folder,"*.py"))

    def ingest_hooks(self):
        if not self.hook_files:
            return
        for pyfile in self.hook_files:
            try:
                self.logger.info("Processing hook file '%s'" % pyfile)
                self.process_hook_file(pyfile)
            except Exception as e:
                self.logger.exception("Can't process hook file: %s" % e)

    def process_hook_file(self, hook_file):
        strcode = open(hook_file).read()
        code = compile(strcode, "<string>", "exec") 
        eval(code, self.shell.extra_ns, self.shell.extra_ns)

    def configure_managers(self):
        if self.managers is not None:
            raise Exception("Managers have already been configured")
        self.managers = {}

        self.logger.info("Setting up managers for following features: %s",
                         self.features)
        assert "job" in self.features, "'job' feature is mandatory"
        if "source" in self.features:
            assert "dump" in self.features and "upload" in self.features, "'source' feature requires both 'dump' and 'upload' features"
        if "dataplugin" in self.features:
            assert "source" in self.features, "'dataplugin' feature requires 'source' feature"

        # specific order, eg. job_manager is used by all managers
        for feat in self.features:
            if hasattr(self, "configure_%s_manager" % feat):
                self.logger.info("Configuring feature '%s'", feat)
                getattr(self, "configure_%s_manager" % feat)()
                self.remaining_features.remove(feat)
            elif hasattr(self, "configure_%s_feature" % feat):
                # see configure_remaining_features()
                pass  # this is configured after managers but should not produce an error
            else:
                raise AttributeError(
                    "Feature '%s' listed but no 'configure_%s_{manager|feature}' method found"
                    % (feat, feat))

        self.logger.info("Active manager(s): %s" % pformat(self.managers))

    def configure_config_feature(self):
        # just a placeholder
        pass

    def configure_upgrade_feature(self):
        """
        Allows a Hub to check for new versions (new commits to apply on running branch)
        and apply them on current code base
        """

        if not getattr(config, "app_folder", None) or not getattr(config, "biothings_folder", None):
            self.logger.warning("Can't schedule check for new code updates, "
                                + "app folder and/or biothings folder not defined")
            return

        from biothings.hub.upgrade import BioThingsSystemUpgrade, ApplicationSystemUpgrade
        def get_upgrader(klass, folder):
            version = get_version(folder)
            klass.SRC_ROOT_FOLDER = folder
            klass.GIT_REPO_URL = version["giturl"]
            klass.DEFAULT_BRANCH = version["branch"]
            return klass

        bt_upgrader_class = get_upgrader(BioThingsSystemUpgrade, config.biothings_folder)
        app_upgrader_class = get_upgrader(ApplicationSystemUpgrade, config.app_folder)
        self.managers["dump_manager"].register_classes([bt_upgrader_class, app_upgrader_class])

        @asyncio.coroutine
        def check_code_upgrade():
            self.logger.info("Checking for new code updates")
            bt_new = check_new_version(config.biothings_folder)
            try:
                app_new = check_new_version(config.app_folder)
            except Exception as e:
                self.logger.warning("Can't check for new version: %s" % e)
                return
            # enrich existing version information with an "upgrade" field.
            # note: we do that on config.conf, the actual config.py module,
            # *not* directly on config as it's a wrapper over config.conf
            for (name, new, param) in (("app", app_new, "APP_VERSION"), ("biothings", bt_new, "BIOTHINGS_VERSION")):
                if new:
                    self.logger.info("Found updates for %s:\n%s" % (name, pformat(new)))
                    getattr(config.conf, param)["upgrade"] = new
                else:
                    # just in case, we pop out the key
                    getattr(config.conf, param).pop("upgrade", None)

        loop = self.managers.get("job_manager") and self.managers[
            "job_manager"].loop or asyncio.get_event_loop()

        # check at startup, then regularly
        asyncio.ensure_future(check_code_upgrade())
        aiocron.crontab(HUB_CHECK_UPGRADE,
                        func=check_code_upgrade,
                        start=True,
                        loop=loop)

    def get_websocket_urls(self):

        if self.ws_urls:
            return self.ws_urls

        import sockjs.tornado
        import biothings.hub.api.handlers.ws as ws
        from biothings.utils.hub_db import ChangeWatcher
        # monitor change in database to report activity in webapp
        self.db_listener = ws.HubDBListener()
        ChangeWatcher.add(self.db_listener)
        ChangeWatcher.publish()
        self.log_listener = ws.LogListener()
        # push log statements to the webapp
        root_logger = logging.getLogger(
        )  # careful, asyncio logger will trigger log statement while in the handler
        # (ie. infinite loop), root logger not recommended)
        root_logger.addHandler(WSLogHandler(self.log_listener))
        self.ws_listeners.extend([self.db_listener, self.log_listener])
        ws_router = sockjs.tornado.SockJSRouter(
            partial(ws.WebSocketConnection, listeners=self.ws_listeners),
            '/ws')
        self.ws_urls = ws_router.urls
        return self.ws_urls

    def configure_ws_feature(self):
        # add websocket endpoint
        ws_urls = self.get_websocket_urls()
        self.routes.extend(ws_urls)

    def configure_terminal_feature(self):
        assert "ws" in self.features, "'terminal' feature requires 'ws'"
        assert "ws" in self.remaining_features, "'terminal' feature should configured before 'ws'"
        # shell logger/listener to communicate between webapp and hub ssh console
        import biothings.hub.api.handlers.ws as ws
        shell_listener = ws.LogListener()
        shell_logger = logging.getLogger("shell")
        assert isinstance(shell_logger,
                          ShellLogger), "shell_logger isn't properly set"
        shell_logger.addHandler(WSShellHandler(shell_listener))
        self.ws_listeners.append(shell_listener)
        # webapp terminal to hub shell connection through /shell endpoint
        from biothings.hub.api.handlers.shell import ShellHandler
        shell_endpoint = ("/shell", ShellHandler, {
            "shell": self.shell,
            "shellog": shell_logger
        })
        self.routes.append(shell_endpoint)

    def configure_dataupload_feature(self):
        assert "ws" in self.features, "'dataupload' feature requires 'ws'"
        assert "ws" in self.remaining_features, "'dataupload' feature should configured before 'ws'"
        # this one is not bound to a specific command
        from biothings.hub.api.handlers.upload import UploadHandler
        # tuple type = interpreted as a route handler
        self.routes.append(
            (r"/dataupload/([\w\.-]+)?", UploadHandler, self.dataupload_config))

    def configure_reloader_feature(self):
        monitored_folders = self.reloader_config["folders"] or [
            "hub/dataload/sources",
            getattr(config, "DATA_PLUGIN_FOLDER", None),
            getattr(config, "HOOKS_FOLDER", "./hooks"),
        ]
        reload_func = self.reloader_config["reload_func"] or partial(
            self.shell.restart, force=True)
        reloader = get_hub_reloader(monitored_folders,
                                    reload_func=reload_func)
        reloader and reloader.monitor()

    def configure_readonly_feature(self):
        """
        Define then expose read-only Hub API endpoints
        so Hub can be accessed without any risk of modifying data
        """
        assert self.api_config is not False, "api_config (read/write API) is required " \
                                             + "to defined a read-only API (it's derived from)"
        # first websockets URLs (we only fetch data from websocket, so no
        # risk of write operations there
        ws_urls = self.get_websocket_urls()
        self.readonly_routes.extend(ws_urls)
        # the rest of the readonly feature setup is done as the end, when starting the server

    def configure_remaining_features(self):
        self.logger.info("Setting up remaining features: %s",
                         self.remaining_features)
        # specific order, eg. job_manager is used by all managers
        for feat in copy.deepcopy(self.remaining_features):
            if hasattr(self, "configure_%s_feature" % feat):
                getattr(self, "configure_%s_feature" % feat)()
                self.remaining_features.remove(feat)
                pass  # this is configured after managers but should not produce an error
            else:
                raise AttributeError(
                    "Feature '%s' listed but no 'configure_%s_feature' method found"
                    % (feat, feat))

    def configure_commands(self):
        """
        Configure hub commands according to available managers
        """
        assert self.managers, "No managers configured"
        self.commands = HubCommands()
        self.commands["status"] = CommandDefinition(command=partial(status, self.managers),
                                                    tracked=False)
        if "config" in self.features:
            self.commands["config"] = CommandDefinition(command=config.show,
                                                        tracked=False)
            self.commands["setconf"] = config.store_value_to_db
            self.commands["resetconf"] = config.reset
        # getting info
        if self.managers.get("source_manager"):
            self.commands["source_info"] = CommandDefinition(
                command=self.managers["source_manager"].get_source,
                tracked=False)
            self.commands["source_reset"] = CommandDefinition(
                command=self.managers["source_manager"].reset, tracked=True)
        # dump commands
        if self.managers.get("dump_manager"):
            self.commands["dump"] = self.managers["dump_manager"].dump_src
            self.commands["dump_all"] = self.managers["dump_manager"].dump_all
        # upload commands
        if self.managers.get("upload_manager"):
            self.commands["upload"] = self.managers[
                "upload_manager"].upload_src
            self.commands["upload_all"] = self.managers[
                "upload_manager"].upload_all
        # building/merging
        if self.managers.get("build_manager"):
            self.commands["whatsnew"] = CommandDefinition(
                command=self.managers["build_manager"].whatsnew, tracked=False)
            self.commands["lsmerge"] = self.managers[
                "build_manager"].list_merge
            self.commands["rmmerge"] = self.managers[
                "build_manager"].delete_merge
            self.commands["merge"] = self.managers["build_manager"].merge
            self.commands["archive"] = self.managers[
                "build_manager"].archive_merge
        if hasattr(config, "INDEX_CONFIG"):
            self.commands["index_config"] = config.INDEX_CONFIG
        if hasattr(config, "SNAPSHOT_CONFIG"):
            self.commands["snapshot_config"] = config.SNAPSHOT_CONFIG
        if hasattr(config, "PUBLISH_CONFIG"):
            self.commands["publish_config"] = config.PUBLISH_CONFIG
        # diff
        if self.managers.get("diff_manager"):
            self.commands["diff"] = self.managers["diff_manager"].diff
            self.commands["report"] = self.managers["diff_manager"].diff_report
        # indexing commands
        if self.managers.get("index_manager"):
            self.commands["index"] = self.managers["index_manager"].index
        if self.managers.get("snapshot_manager"):
            self.commands["snapshot"] = self.managers[
                "snapshot_manager"].snapshot
        # data release commands
        if self.managers.get("release_manager"):
            self.commands["create_release_note"] = self.managers[
                "release_manager"].create_release_note
            self.commands["get_release_note"] = CommandDefinition(
                command=self.managers["release_manager"].get_release_note,
                tracked=False)
            self.commands["publish"] = self.managers["release_manager"].publish
            self.commands["publish_diff"] = self.managers[
                "release_manager"].publish_diff
            self.commands["publish_snapshot"] = self.managers[
                "release_manager"].publish_snapshot
        if self.managers.get("sync_manager"):
            self.commands["sync"] = CommandDefinition(
                command=self.managers["sync_manager"].sync)
        # inspector
        if self.managers.get("inspect_manager"):
            self.commands["inspect"] = self.managers["inspect_manager"].inspect
        # data plugins
        if self.managers.get("assistant_manager"):
            self.commands["register_url"] = partial(
                self.managers["assistant_manager"].register_url)
            self.commands["unregister_url"] = partial(
                self.managers["assistant_manager"].unregister_url)
            self.commands["export_plugin"] = partial(
                self.managers["assistant_manager"].export)
        if self.managers.get("dataplugin_manager"):
            self.commands["dump_plugin"] = self.managers[
                "dataplugin_manager"].dump_src
        if "autohub" in self.DEFAULT_FEATURES:
            self.commands["list"] = CommandDefinition(command=self.autohub_feature.list_biothings, tracked=False)
            # dump commands
            self.commands["versions"] = partial(self.managers["dump_manager"].call, method_name="versions")
            self.commands["check"] = partial(self.managers["dump_manager"].dump_src, check_only=True)
            self.commands["info"] = partial(self.managers["dump_manager"].call, method_name="info")
            self.commands["download"] = partial(self.managers["dump_manager"].dump_src)
            # upload commands
            self.commands["apply"] = partial(self.managers["upload_manager"].upload_src)
            self.commands["install"] = partial(self.autohub_feature.install)
            self.commands["backend"] = partial(self.managers["dump_manager"].call, method_name="get_target_backend")
            self.commands["reset_backend"] = partial(self.managers["dump_manager"].call, method_name="reset_target_backend")

        logging.info("Registered commands: %s", list(self.commands.keys()))

    def configure_extra_commands(self):
        """
        Same as configure_commands() but commands are not exposed publicly in the shell
        (they are shortcuts or commands for API endpoints, supporting commands, etc...)
        """
        assert self.managers, "No managers configured"
        self.extra_commands = {}  # unordered since not exposed, we don't care
        loop = self.managers.get("job_manager") and self.managers[
            "job_manager"].loop or asyncio.get_event_loop()
        self.extra_commands["g"] = CommandDefinition(command=globals(),
                                                     tracked=False)
        self.extra_commands["sch"] = CommandDefinition(command=partial(get_schedule, loop),
                                                       tracked=False)
        # expose contant so no need to put quotes (eg. top(pending) instead of top("pending")
        self.extra_commands["pending"] = CommandDefinition(command=pending,
                                                           tracked=False)
        self.extra_commands["loop"] = CommandDefinition(command=loop,
                                                        tracked=False)

        if self.managers.get("job_manager"):
            self.extra_commands["pqueue"] = CommandDefinition(
                command=self.managers["job_manager"].process_queue,
                tracked=False)
            self.extra_commands["tqueue"] = CommandDefinition(
                command=self.managers["job_manager"].thread_queue,
                tracked=False)
            self.extra_commands["jm"] = CommandDefinition(
                command=self.managers["job_manager"], tracked=False)
            self.extra_commands["top"] = CommandDefinition(
                command=self.managers["job_manager"].top, tracked=False)
            self.extra_commands["job_info"] = CommandDefinition(
                command=self.managers["job_manager"].job_info, tracked=False)
            self.extra_commands["schedule"] = CommandDefinition(
                command=self.managers["job_manager"].schedule, tracked=False)
        if self.managers.get("source_manager"):
            self.extra_commands["sm"] = CommandDefinition(
                command=self.managers["source_manager"], tracked=False)
            self.extra_commands["sources"] = CommandDefinition(
                command=self.managers["source_manager"].get_sources,
                tracked=False)
            self.extra_commands["source_save_mapping"] = CommandDefinition(
                command=self.managers["source_manager"].save_mapping)
        if self.managers.get("dump_manager"):
            self.extra_commands["dm"] = CommandDefinition(
                command=self.managers["dump_manager"], tracked=False)
            self.extra_commands["dump_info"] = CommandDefinition(
                command=self.managers["dump_manager"].dump_info, tracked=False)
        if self.managers.get("dataplugin_manager"):
            self.extra_commands["dpm"] = CommandDefinition(
                command=self.managers["dataplugin_manager"], tracked=False)
        if self.managers.get("assistant_manager"):
            self.extra_commands["am"] = CommandDefinition(
                command=self.managers["assistant_manager"], tracked=False)
        if self.managers.get("upload_manager"):
            self.extra_commands["um"] = CommandDefinition(
                command=self.managers["upload_manager"], tracked=False)
            self.extra_commands["upload_info"] = CommandDefinition(
                command=self.managers["upload_manager"].upload_info,
                tracked=False)
        if self.managers.get("build_manager"):
            self.extra_commands["bm"] = CommandDefinition(
                command=self.managers["build_manager"], tracked=False)
            self.extra_commands["builds"] = CommandDefinition(
                command=self.managers["build_manager"].build_info,
                tracked=False)
            self.extra_commands["build"] = CommandDefinition(
                command=lambda id: self.managers["build_manager"].build_info(
                    id=id),
                tracked=False)
            self.extra_commands["build_config_info"] = CommandDefinition(
                command=self.managers["build_manager"].build_config_info,
                tracked=False)
            self.extra_commands["build_save_mapping"] = CommandDefinition(
                command=self.managers["build_manager"].save_mapping)
            self.extra_commands["create_build_conf"] = CommandDefinition(
                command=self.managers["build_manager"].
                create_build_configuration)
            self.extra_commands["update_build_conf"] = CommandDefinition(
                command=self.managers["build_manager"].
                update_build_configuration)
            self.extra_commands["delete_build_conf"] = CommandDefinition(
                command=self.managers["build_manager"].
                delete_build_configuration)
        if self.managers.get("diff_manager"):
            self.extra_commands["dim"] = CommandDefinition(
                command=self.managers["diff_manager"], tracked=False)
            self.extra_commands["diff_info"] = CommandDefinition(
                command=self.managers["diff_manager"].diff_info, tracked=False)
            self.extra_commands["jsondiff"] = CommandDefinition(
                command=jsondiff, tracked=False)
        if self.managers.get("sync_manager"):
            self.extra_commands["sym"] = CommandDefinition(
                command=self.managers["sync_manager"], tracked=False)
        if self.managers.get("index_manager"):
            self.extra_commands["im"] = CommandDefinition(
                command=self.managers["index_manager"], tracked=False)
            self.extra_commands["index_info"] = CommandDefinition(
                command=self.managers["index_manager"].index_info,
                tracked=False)
            self.extra_commands["validate_mapping"] = CommandDefinition(
                command=self.managers["index_manager"].validate_mapping)
        if self.managers.get("snapshot_manager"):
            self.extra_commands["ssm"] = CommandDefinition(
                command=self.managers["snapshot_manager"], tracked=False)
            self.extra_commands["snapshot_info"] = CommandDefinition(
                command=self.managers["snapshot_manager"].snapshot_info,
                tracked=False)
        if self.managers.get("release_manager"):
            self.extra_commands["rm"] = CommandDefinition(
                command=self.managers["release_manager"], tracked=False)
            self.extra_commands["release_info"] = CommandDefinition(
                command=self.managers["release_manager"].release_info,
                tracked=False)
            self.extra_commands["reset_synced"] = CommandDefinition(
                command=self.managers["release_manager"].reset_synced,
                tracked=True)
        if self.managers.get("inspect_manager"):
            self.extra_commands["ism"] = CommandDefinition(
                command=self.managers["inspect_manager"], tracked=False)
        if self.managers.get("api_manager"):
            self.extra_commands["api"] = CommandDefinition(
                command=self.managers["api_manager"], tracked=False)
            self.extra_commands["get_apis"] = CommandDefinition(
                command=self.managers["api_manager"].get_apis, tracked=False)
            self.extra_commands["delete_api"] = CommandDefinition(
                command=self.managers["api_manager"].delete_api)
            self.extra_commands["create_api"] = CommandDefinition(
                command=self.managers["api_manager"].create_api)
            self.extra_commands["start_api"] = CommandDefinition(
                command=self.managers["api_manager"].start_api)
            self.extra_commands["stop_api"] = self.managers[
                "api_manager"].stop_api
        if "upgrade" in self.DEFAULT_FEATURES:
            def upgrade(code_base):  # just a wrapper over dumper
                """Upgrade (git pull) repository for given code base name ("biothings_sdk" or "application")"""
                assert code_base in ("application","biothings_sdk"), "Unknown code base '%s'" % code_base
                return self.managers["dump_manager"].dump_src("__" + code_base)
            self.commands["upgrade"] = CommandDefinition(command=upgrade)

        self.extra_commands["expose"] = self.add_api_endpoint

        logging.debug("Registered extra (private) commands: %s",
                      list(self.extra_commands.keys()))

    def add_api_endpoint(self, endpoint_name, command_name, method, **kwargs):
        """
        Add an API endpoint to expose command named "command_name"
        using HTTP method "method". **kwargs are used to specify
        more arguments for EndpointDefinition
        """
        if self.configured:
            raise Exception("API endpoint creation must be done before Hub is configured")
        from biothings.hub.api import EndpointDefinition
        endpoint = EndpointDefinition(name=command_name,method=method,**kwargs)
        self.api_endpoints[endpoint_name] = endpoint

    def configure_api_endpoints(self):
        cmdnames = list(self.commands.keys())
        if self.extra_commands:
            cmdnames.extend(list(self.extra_commands.keys()))
        from biothings.hub.api import EndpointDefinition
        self.api_endpoints["config"] = []
        if "config" in cmdnames:
            self.api_endpoints["config"].append(
                EndpointDefinition(name="config", method="get"))
            self.api_endpoints["config"].append(
                EndpointDefinition(name="setconf",
                                   method="put",
                                   force_bodyargs=True))
            self.api_endpoints["config"].append(
                EndpointDefinition(name="resetconf",
                                   method="delete",
                                   force_bodyargs=True))
        if not self.api_endpoints["config"]:
            self.api_endpoints.pop("config")
        if "builds" in cmdnames:
            self.api_endpoints["builds"] = EndpointDefinition(name="builds",
                                                              method="get")
        self.api_endpoints["build"] = []
        if "build" in cmdnames:
            self.api_endpoints["build"].append(
                EndpointDefinition(method="get", name="build"))
        if "archive" in cmdnames:
            self.api_endpoints["build"].append(
                EndpointDefinition(method="post",
                                   name="archive",
                                   suffix="archive"))
        if "rmmerge" in cmdnames:
            self.api_endpoints["build"].append(
                EndpointDefinition(method="delete", name="rmmerge"))
        if "merge" in cmdnames:
            self.api_endpoints["build"].append(
                EndpointDefinition(name="merge", method="put", suffix="new"))
        if "build_save_mapping" in cmdnames:
            self.api_endpoints["build"].append(
                EndpointDefinition(name="build_save_mapping",
                                   method="put",
                                   suffix="mapping"))
        if not self.api_endpoints["build"]:
            self.api_endpoints.pop("build")
        self.api_endpoints["publish"] = []
        if "publish_diff" in cmdnames:
            self.api_endpoints["publish"].append(
                EndpointDefinition(name="publish_diff",
                                   method="post",
                                   suffix="incremental",
                                   force_bodyargs=True))
        if "publish_snapshot" in cmdnames:
            self.api_endpoints["publish"].append(
                EndpointDefinition(name="publish_snapshot",
                                   method="post",
                                   suffix="full",
                                   force_bodyargs=True))
        if not self.api_endpoints["publish"]:
            self.api_endpoints.pop("publish")
        if "diff" in cmdnames:
            self.api_endpoints["diff"] = EndpointDefinition(
                name="diff", method="put", force_bodyargs=True)
        if "job_info" in cmdnames:
            self.api_endpoints["job_manager"] = EndpointDefinition(
                name="job_info", method="get")
        if "dump_info" in cmdnames:
            self.api_endpoints["dump_manager"] = EndpointDefinition(
                name="dump_info", method="get")
        if "upload_info" in cmdnames:
            self.api_endpoints["upload_manager"] = EndpointDefinition(
                name="upload_info", method="get")
        if "build_config_info" in cmdnames:
            self.api_endpoints["build_manager"] = EndpointDefinition(
                name="build_config_info", method="get")
        if "index_info" in cmdnames:
            self.api_endpoints["index_manager"] = EndpointDefinition(
                name="index_info", method="get")
        if "snapshot_info" in cmdnames:
            self.api_endpoints["snapshot_manager"] = EndpointDefinition(
                name="snapshot_info", method="get")
        if "release_info" in cmdnames:
            self.api_endpoints["release_manager"] = EndpointDefinition(
                name="release_info", method="get")
        if "reset_synced" in cmdnames:
            self.api_endpoints[
                "release_manager/reset_synced"] = EndpointDefinition(
                    name="reset_synced", method="put")
        if "diff_info" in cmdnames:
            self.api_endpoints["diff_manager"] = EndpointDefinition(
                name="diff_info", method="get")
        if "commands" in cmdnames:
            self.api_endpoints["commands"] = EndpointDefinition(
                name="commands", method="get")
        if "command" in cmdnames:
            self.api_endpoints["command"] = EndpointDefinition(name="command",
                                                               method="get")
        if "sources" in cmdnames:
            self.api_endpoints["sources"] = EndpointDefinition(name="sources",
                                                               method="get")
        self.api_endpoints["source"] = []
        if "source_info" in cmdnames:
            self.api_endpoints["source"].append(
                EndpointDefinition(name="source_info", method="get"))
        if "source_reset" in cmdnames:
            self.api_endpoints["source"].append(
                EndpointDefinition(name="source_reset",
                                   method="post",
                                   suffix="reset"))
        if "dump" in cmdnames:
            self.api_endpoints["source"].append(
                EndpointDefinition(name="dump", method="put", suffix="dump"))
        if "upload" in cmdnames:
            self.api_endpoints["source"].append(
                EndpointDefinition(name="upload",
                                   method="put",
                                   suffix="upload"))
        if "source_save_mapping" in cmdnames:
            self.api_endpoints["source"].append(
                EndpointDefinition(name="source_save_mapping",
                                   method="put",
                                   suffix="mapping"))
        if not self.api_endpoints["source"]:
            self.api_endpoints.pop("source")
        if "inspect" in cmdnames:
            self.api_endpoints["inspect"] = EndpointDefinition(
                name="inspect", method="put", force_bodyargs=True)
        if "register_url" in cmdnames:
            self.api_endpoints["dataplugin/register_url"] = EndpointDefinition(
                name="register_url", method="post", force_bodyargs=True)
        if "unregister_url" in cmdnames:
            self.api_endpoints[
                "dataplugin/unregister_url"] = EndpointDefinition(
                    name="unregister_url",
                    method="delete",
                    force_bodyargs=True)
        self.api_endpoints["dataplugin"] = []
        if "dump_plugin" in cmdnames:
            self.api_endpoints["dataplugin"].append(
                EndpointDefinition(name="dump_plugin",
                                   method="put",
                                   suffix="dump"))
        if "export_plugin" in cmdnames:
            self.api_endpoints["dataplugin"].append(
                EndpointDefinition(name="export_plugin",
                                   method="put",
                                   suffix="export"))
        if not self.api_endpoints["dataplugin"]:
            self.api_endpoints.pop("dataplugin")
        if "jsondiff" in cmdnames:
            self.api_endpoints["jsondiff"] = EndpointDefinition(
                name="jsondiff", method="post", force_bodyargs=True)
        if "validate_mapping" in cmdnames:
            self.api_endpoints["mapping/validate"] = EndpointDefinition(
                name="validate_mapping", method="post", force_bodyargs=True)
        self.api_endpoints["buildconf"] = []
        if "create_build_conf" in cmdnames:
            self.api_endpoints["buildconf"].append(
                EndpointDefinition(name="create_build_conf",
                                   method="post",
                                   force_bodyargs=True))
            self.api_endpoints["buildconf"].append(
                EndpointDefinition(name="update_build_conf",
                                   method="put",
                                   force_bodyargs=True))
        if "delete_build_conf" in cmdnames:
            self.api_endpoints["buildconf"].append(
                EndpointDefinition(name="delete_build_conf",
                                   method="delete",
                                   force_bodyargs=True))
        if not self.api_endpoints["buildconf"]:
            self.api_endpoints.pop("buildconf")
        if "index" in cmdnames:
            self.api_endpoints["index"] = EndpointDefinition(
                name="index", method="put", force_bodyargs=True)
        if "snapshot" in cmdnames:
            self.api_endpoints["snapshot"] = EndpointDefinition(
                name="snapshot", method="put", force_bodyargs=True)
        if "sync" in cmdnames:
            self.api_endpoints["sync"] = EndpointDefinition(
                name="sync", method="post", force_bodyargs=True)
        if "whatsnew" in cmdnames:
            self.api_endpoints["whatsnew"] = EndpointDefinition(
                name="whatsnew", method="get")
        if "status" in cmdnames:
            self.api_endpoints["status"] = EndpointDefinition(name="status",
                                                              method="get")
        self.api_endpoints["release_note"] = []
        if "create_release_note" in cmdnames:
            self.api_endpoints["release_note"].append(
                EndpointDefinition(name="create_release_note",
                                   method="put",
                                   suffix="create",
                                   force_bodyargs=True))
        if "get_release_note" in cmdnames:
            self.api_endpoints["release_note"].append(
                EndpointDefinition(name="get_release_note",
                                   method="get",
                                   force_bodyargs=True))
        if not self.api_endpoints["release_note"]:
            self.api_endpoints.pop("release_note")
        self.api_endpoints["api"] = []
        if "start_api" in cmdnames:
            self.api_endpoints["api"].append(
                EndpointDefinition(name="start_api",
                                   method="put",
                                   suffix="start"))
        if "stop_api" in cmdnames:
            self.api_endpoints["api"].append(
                EndpointDefinition(name="stop_api",
                                   method="put",
                                   suffix="stop"))
        if "delete_api" in cmdnames:
            self.api_endpoints["api"].append(
                EndpointDefinition(name="delete_api",
                                   method="delete",
                                   force_bodyargs=True))
        if "create_api" in cmdnames:
            self.api_endpoints["api"].append(
                EndpointDefinition(name="create_api",
                                   method="post",
                                   force_bodyargs=True))
        if not self.api_endpoints["api"]:
            self.api_endpoints.pop("api")
        if "get_apis" in cmdnames:
            self.api_endpoints["api/list"] = EndpointDefinition(
                name="get_apis", method="get")
        if "stop" in cmdnames:
            self.api_endpoints["stop"] = EndpointDefinition(name="stop",
                                                            method="put")
        if "restart" in cmdnames:
            self.api_endpoints["restart"] = EndpointDefinition(name="restart",
                                                               method="put")
        self.api_endpoints["standalone"] = []
        if "list" in cmdnames:
            self.api_endpoints["standalone"].append(EndpointDefinition(name="list", method="get", suffix="list"))
        if "versions" in cmdnames:
            self.api_endpoints["standalone"].append(EndpointDefinition(name="versions", method="get", suffix="versions"))
        if "check" in cmdnames:
            self.api_endpoints["standalone"].append(EndpointDefinition(name="check", method="get", suffix="check"))
        if "info" in cmdnames:
            self.api_endpoints["standalone"].append(EndpointDefinition(name="info", method="get", suffix="info"))
        if "download" in cmdnames:
            self.api_endpoints["standalone"].append(EndpointDefinition(name="download", method="post", suffix="download"))
        if "apply" in cmdnames:
            self.api_endpoints["standalone"].append(EndpointDefinition(name="apply", method="post", suffix="apply"))
        if "install" in cmdnames:
            self.api_endpoints["standalone"].append(EndpointDefinition(name="install", method="post", suffix="install"))
        if "backend" in cmdnames:
            self.api_endpoints["standalone"].append(EndpointDefinition(name="backend", method="get", suffix="backend"))
        if "reset_backend" in cmdnames:
            self.api_endpoints["standalone"].append(EndpointDefinition(name="reset_backend", method="delete", suffix="backend"))
        if not self.api_endpoints["standalone"]:
            self.api_endpoints.pop("standalone")
        if "upgrade" in self.commands:
            self.api_endpoints["code/upgrade"] = EndpointDefinition(name="upgrade", method="put")


class HubSSHServer(asyncssh.SSHServer):

    PASSWORDS = {}
    SHELL = None

    def session_requested(self):
        return HubSSHServerSession(self.__class__.NAME, self.__class__.SHELL)

    def connection_made(self, connection):
        self._conn = connection
        print('SSH connection received from %s.' %
              connection.get_extra_info('peername')[0])

    def connection_lost(self, exc):
        if exc:
            print('SSH connection error: ' + str(exc), file=sys.stderr)
        else:
            print('SSH connection closed.')

    def begin_auth(self, username):
        try:
            self._conn.set_authorized_keys('bin/authorized_keys/%s.pub' %
                                           username)
        except IOError:
            pass
        return True

    def password_auth_supported(self):
        return True

    def validate_password(self, username, password):
        if self.password_auth_supported():
            pw = self.__class__.PASSWORDS.get(username, '*')
            return crypt.crypt(password, pw) == pw
        else:
            return False


class HubSSHServerSession(asyncssh.SSHServerSession):
    def __init__(self, name, shell):
        self.name = name
        self.shell = shell
        self._input = ''

    def connection_made(self, chan):
        self._chan = chan

    def shell_requested(self):
        return True

    def exec_requested(self, command):
        self.eval_lines(["%s" % command, "\n"])
        return True

    def session_started(self):
        welcome = ('\nWelcome to %s, %s!\n' %
                   (self.name, self._chan.get_extra_info('username')))
        self.shell.shellog.output(welcome)
        self._chan.write(welcome)
        prompt = 'hub> '
        self.shell.shellog.output(prompt)
        self._chan.write(prompt)

    def data_received(self, data, datatype):
        self._input += data
        return self.eval_lines(self._input.split('\n'))

    def eval_lines(self, lines):
        for line in lines[:-1]:
            try:
                outs = [out for out in self.shell.eval(line) if out]
                # trailing \n if not already there
                if outs:
                    strout = "\n".join(outs).strip("\n") + "\n"
                    self._chan.write(strout)
                    self.shell.shellog.output(strout)
            except AlreadyRunningException as e:
                self._chan.write("AlreadyRunningException: %s" % e)
            except CommandError as e:
                self._chan.write("CommandError: %s" % e)
        self._chan.write('hub> ')
        # consume passed commands
        self._input = lines[-1]

    def eof_received(self):
        self._chan.write('Have a good one...\n')
        self._chan.exit(0)

    def break_received(self, msec):
        # simulate CR
        self._chan.write('\n')
        self.data_received("\n", None)



import json

import requests


def slack_msg(webhook, msg, color='warning'):
    headers = {'content-type': 'application/json'}
    _msg = msg.lower()
    for keyword in ['fail', 'error']:
        if _msg.find(keyword) != -1:
            color = 'danger'
            break
    # we use attachments so we can add colors
    params = {
        "attachments": [{
            "text": msg,
            "color" : color,
        }]
    }
    res = requests.post(webhook, json.dumps(params), headers=headers)
    res.raise_for_status()



import copy
import itertools
import json
import re
import time

from elasticsearch import (Elasticsearch, NotFoundError, RequestError,
                           TransportError, ElasticsearchException)
from elasticsearch import helpers

from biothings.utils.common import iter_n, splitstr, nan, inf
from biothings.utils.dataload import dict_walk

# setup ES logging
import logging
formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
es_logger = logging.getLogger('elasticsearch')
es_logger.setLevel(logging.WARNING)
ch = logging.StreamHandler()
ch.setFormatter(formatter)
es_logger.addHandler(ch)

es_tracer = logging.getLogger('elasticsearch.trace')
es_tracer.setLevel(logging.WARNING)
ch = logging.StreamHandler()
ch.setFormatter(formatter)
es_tracer.addHandler(ch)

def verify_ids(doc_iter, es_host, index, doc_type=None, step=100000, ):
    '''verify how many docs from input interator/list overlapping with existing docs.'''

    index = index
    doc_type = doc_type
    es = get_es(es_host)
    q = {'query': {'ids': {"values": []}}}
    total_cnt = 0
    found_cnt = 0
    out = []
    for doc_batch in iter_n(doc_iter, n=step):
        id_li = [doc['_id'] for doc in doc_batch]
        # id_li = [doc['_id'].replace('chr', '') for doc in doc_batch]
        q['query']['ids']['values'] = id_li
        xres = es.search(index=index, doc_type=doc_type, body=q, _source=False)
        found_cnt += xres['hits']['total']
        total_cnt += len(id_li)
        out.extend([x['_id'] for x in xres['hits']['hits']])
    return out


def get_es(es_host, timeout=120, max_retries=3, retry_on_timeout=False):
    es = Elasticsearch(es_host, timeout=timeout, max_retries=max_retries,
                       retry_on_timeout=retry_on_timeout)
    return es


def wrapper(func):
    '''this wrapper allows passing index and doc_type from wrapped method.'''
    def outter_fn(*args, **kwargs):
        self = args[0]
        index = kwargs.pop('index', self._index)             # pylint: disable=protected-access
        doc_type = kwargs.pop('doc_type', self._doc_type)    # pylint: disable=protected-access
        self._index = index                                  # pylint: disable=protected-access
        self._doc_type = doc_type                            # pylint: disable=protected-access
        return func(*args, **kwargs)
    outter_fn.__doc__ = func.__doc__
    return outter_fn


class IndexerException(Exception):
    pass

class ESIndexer():
    def __init__(self, index, doc_type, es_host, step=10000,
                 number_of_shards=10, number_of_replicas=0,
                 check_index=True, **kwargs):
        self.es_host = es_host
        self._es = get_es(es_host, **kwargs)
        if check_index:
            # if index is actually an alias, resolve the alias to
            # the real underlying index
            try:
                res = self._es.indices.get_alias(index)
                # this was an alias
                assert len(res) == 1, "Expecing '%s' to be an alias, but got nothing..." % index
                self._index = list(res.keys())[0]
            except NotFoundError:
                # this was a real index name
                self._index = index
        self._doc_type = None
        if doc_type:
            self._doc_type = doc_type
        else:
            # assuming index exists, get mapping to discover doc_type
            try:
                m = self.get_mapping()
                assert len(m) == 1, "Expected only one doc type, got: %s" % m.keys()
                self._doc_type = list(m).pop()
            except Exception as e:       # pylint: disable=broad-except
                if check_index:
                    logging.info("Failed to guess doc_type: %s", e)
        self.number_of_shards = number_of_shards            # set number_of_shards when create_index
        self.number_of_replicas = int(number_of_replicas)   # set number_of_replicas when create_index
        self.step = step   # the bulk size when doing bulk operation.
        self.s = None      # number of records to skip, useful to continue indexing after an error.

    @wrapper
    def get_biothing(self, bid, only_source=False, **kwargs):
        rawdoc = self._es.get(index=self._index, id=bid, doc_type=self._doc_type, **kwargs)
        if not only_source:
            return rawdoc
        else:
            doc = rawdoc['_source']
            doc["_id"] = rawdoc["_id"]
            return doc

    @wrapper
    def exists(self, bid):
        """return True/False if a biothing id exists or not."""
        try:
            doc = self.get_biothing(bid, stored_fields=None)
            return doc['found']
        except NotFoundError:
            return False

    @wrapper
    def mexists(self, bid_list):
        q = {
            "query": {
                "ids": {
                    "values": bid_list
                }
            }
        }
        res = self._es.search(index=self._index, doc_type=self._doc_type, body=q, stored_fields=None, size=len(bid_list))
        # id_set = set([doc['_id'] for doc in res['hits']['hits']])     # TODO: Confirm this line
        id_set = {doc['_id'] for doc in res['hits']['hits']}
        return [(bid, bid in id_set) for bid in bid_list]

    @wrapper
    def count(self, q=None, raw=False):
        try:
            _res = self._es.count(self._index, self._doc_type, q)
            return _res if raw else _res['count']
        except NotFoundError:
            return None

    @wrapper
    def count_src(self, src):
        if isinstance(src, str):
            src = [src]
        cnt_d = {}
        for _src in src:
            q = {
                "query": {
                    "constant_score": {
                        "filter": {
                            "exists": {"field": _src}
                        }
                    }
                }
            }
            cnt_d[_src] = self.count(q)
        return cnt_d

    @wrapper
    def create_index(self, mapping=None, extra_settings=None):
        if not self._es.indices.exists(self._index):
            body = {
                'settings': {
                    'number_of_shards': self.number_of_shards,
                    "number_of_replicas": self.number_of_replicas,
                }
            }
            extra_settings = extra_settings or {}
            body["settings"].update(extra_settings)
            if mapping:
                mapping = {"mappings": mapping}
                body.update(mapping)
            self._es.indices.create(index=self._index, body=body)

    @wrapper
    def exists_index(self):
        return self._es.indices.exists(self._index)

    def index(self, doc, id=None, action="index"):       # pylint: disable=redefined-builtin
        '''add a doc to the index. If id is not None, the existing doc will be
           updated.
        '''
        self._es.index(self._index, self._doc_type, doc, id=id, params={"op_type": action})

    def index_bulk(self, docs, step=None, action='index'):
        index_name = self._index
        doc_type = self._doc_type
        step = step or self.step

        def _get_bulk(doc):
            # keep original doc
            ndoc = copy.copy(doc)
            ndoc.update({
                "_index": index_name,
                "_type": doc_type,
                "_op_type": action,
            })
            return ndoc
        actions = (_get_bulk(doc) for doc in docs)
        num_ok, errors = helpers.bulk(self._es, actions, chunk_size=step)
        if errors:
            raise ElasticsearchException("%d errors while bulk-indexing: %s" % (len(errors), [str(e) for e in errors]))
        return num_ok, errors

    def delete_doc(self, id):                # pylint: disable=redefined-builtin
        '''delete a doc from the index based on passed id.'''
        return self._es.delete(self._index, self._doc_type, id)

    def delete_docs(self, ids, step=None):
        '''delete a list of docs in bulk.'''
        index_name = self._index
        doc_type = self._doc_type
        step = step or self.step

        def _get_bulk(_id):
            doc = {
                '_op_type': 'delete',
                "_index": index_name,
                "_type": doc_type,
                "_id": _id
            }
            return doc
        actions = (_get_bulk(_id) for _id in ids)
        return helpers.bulk(self._es, actions, chunk_size=step, stats_only=True, raise_on_error=False)

    def delete_index(self):
        self._es.indices.delete(self._index)

    def update(self, id, extra_doc, upsert=True):          # pylint: disable=redefined-builtin
        '''update an existing doc with extra_doc.
           allow to set upsert=True, to insert new docs.
        '''
        body = {'doc': extra_doc}
        if upsert:
            body['doc_as_upsert'] = True
        return self._es.update(self._index, self._doc_type, id, body)

    def update_docs(self, partial_docs, upsert=True, step=None, **kwargs):
        '''update a list of partial_docs in bulk.
           allow to set upsert=True, to insert new docs.
        '''
        index_name = self._index
        doc_type = self._doc_type
        step = step or self.step

        def _get_bulk(doc):
            doc = {
                '_op_type': 'update',
                "_index": index_name,
                "_type": doc_type,
                "_id": doc['_id'],
                "doc": doc
            }
            if upsert:
                doc['doc_as_upsert'] = True
            return doc
        actions = (_get_bulk(doc) for doc in partial_docs)
        return helpers.bulk(self._es, actions, chunk_size=step, **kwargs)

    def get_mapping(self):
        """return the current index mapping"""
        m = self._es.indices.get_mapping(index=self._index, doc_type=self._doc_type)
        return m[self._index]["mappings"]

    def update_mapping(self, m):
        assert list(m) == [self._doc_type], "Bad mapping format, should have one doc_type, got: %s" % list(m)
        assert 'properties' in m[self._doc_type], "Bad mapping format, no 'properties' key"
        return self._es.indices.put_mapping(index=self._index, doc_type=self._doc_type, body=m)

    def get_mapping_meta(self):
        """return the current _meta field."""
        m = self.get_mapping()
        doc_type = self._doc_type
        if doc_type is None:
            # fetch doc_type from mapping
            assert len(m) == 1, "More than one doc_type found, not supported when self._doc_type " + \
                                "is not initialized"
            doc_type = list(m.keys())[0]
        return {"_meta": m[doc_type]["_meta"]}

    def update_mapping_meta(self, meta):
        allowed_keys = {'_meta', '_timestamp'}
        # if isinstance(meta, dict) and len(set(meta) - allowed_keys) == 0:
        if isinstance(meta, dict) and not set(meta) - allowed_keys:
            body = {self._doc_type: meta}
            return self._es.indices.put_mapping(
                doc_type=self._doc_type,
                body=body,
                index=self._index
            )
        else:
            raise ValueError('Input "meta" should have and only have "_meta" field.')

    @wrapper
    def build_index(self, collection, verbose=True, query=None, bulk=True, update=False, allow_upsert=True):
        index_name = self._index
        # update some settings for bulk indexing
        body = {
            "index": {
                "refresh_interval": "-1",              # disable refresh temporarily
                "auto_expand_replicas": "0-all",
            }
        }
        self._es.indices.put_settings(body, index_name)
        try:
            self._build_index_sequential(collection, verbose, query=query, bulk=bulk, update=update, allow_upsert=True)
        finally:
            # restore some settings after bulk indexing is done.
            body = {
                "index": {
                    "refresh_interval": "1s"              # default settings
                }
            }
            self._es.indices.put_settings(body, index_name)

            try:
                self._es.indices.flush()
                self._es.indices.refresh()
            except:          # pylint: disable=bare-except  # noqa
                pass

            time.sleep(1)
            src_cnt = collection.count(query)
            es_cnt = self.count()
            if src_cnt != es_cnt:
                raise IndexerException("Total count of documents does not match [{}, should be {}]".format(es_cnt, src_cnt))

            return es_cnt

    def _build_index_sequential(self, collection, verbose=False, query=None, bulk=True, update=False, allow_upsert=True):

        def rate_control(cnt, t):
            delay = 0
            if t > 90:
                delay = 30
            elif t > 60:
                delay = 10
            if delay:
                time.sleep(delay)

        from biothings.utils.mongo import doc_feeder
        src_docs = doc_feeder(collection, step=self.step, s=self.s, batch_callback=rate_control, query=query)
        if bulk:
            if update:
                # input doc will update existing one
                # if allow_upsert, create new one if not exist
                res = self.update_docs(src_docs, upsert=allow_upsert)
            else:
                # input doc will overwrite existing one
                res = self.index_bulk(src_docs)
            # if len(res[1]) > 0:
            if res[1]:
                raise IndexerException("Error: {} docs failed indexing.".format(len(res[1])))
            return res[0]

        else:
            cnt = 0
            for doc in src_docs:
                self.index(doc)
                cnt += 1
            return cnt

    @wrapper
    def optimize(self, max_num_segments=1):
        '''optimize the default index.'''
        params = {
            "wait_for_merge": False,
            "max_num_segments": max_num_segments,
        }
        return self._es.indices.forcemerge(index=self._index, params=params)

    def clean_field(self, field, dryrun=True, step=5000):
        '''remove a top-level field from ES index, if the field is the only field of the doc,
           remove the doc as well.
           step is the size of bulk update on ES
           try first with dryrun turned on, and then perform the actual updates with dryrun off.
        '''
        q = {
            "query": {
                "constant_score": {
                    "filter": {
                        "exists": {
                            "field": field
                        }
                    }
                }
            }
        }
        cnt_orphan_doc = 0
        cnt = 0
        _li = []
        for doc in self.doc_feeder(query=q):
            if set(doc) == {'_id', field}:
                cnt_orphan_doc += 1
                # delete orphan doc
                _li.append({
                    "delete": {
                        "_index": self._index,
                        "_type": self._doc_type,
                        "_id": doc['_id']
                    }
                })
            else:
                # otherwise, just remove the field from the doc
                _li.append({
                    "update": {
                        "_index": self._index,
                        "_type": self._doc_type,
                        "_id": doc['_id']
                    }
                })
                # this script update requires "script.disable_dynamic: false" setting
                # in elasticsearch.yml
                _li.append({"script": 'ctx._source.remove("{}")'.format(field)})

            cnt += 1
            if len(_li) == step:
                if not dryrun:
                    self._es.bulk(body=_li)
                _li = []
        if _li:
            if not dryrun:
                self._es.bulk(body=_li)

        return {"total": cnt, "updated": cnt - cnt_orphan_doc, "deleted": cnt_orphan_doc}

    @wrapper
    def doc_feeder_using_helper(self, step=None, verbose=True, query=None, scroll='10m', **kwargs):
        # verbose unimplemented
        step = step or self.step
        q = query if query else {'query': {'match_all': {}}}
        for rawdoc in helpers.scan(client=self._es, query=q, scroll=scroll, index=self._index,
                                   doc_type=self._doc_type, **kwargs):
            if rawdoc.get('_source', False):
                doc = rawdoc['_source']
                doc["_id"] = rawdoc["_id"]
                yield doc
            else:
                yield rawdoc

    @wrapper
    def doc_feeder(self, step=None, verbose=True, query=None, scroll='10m', only_source=True, **kwargs):
        step = step or self.step
        q = query if query else {'query': {'match_all': {}}}
        _q_cnt = self.count(q=q, raw=True)
        n = _q_cnt['count']
        n_shards = _q_cnt['_shards']['total']
        assert n_shards == _q_cnt['_shards']['successful']
        # Not sure if scroll size is per shard anymore in the new ES...should check this
        _size = int(step / n_shards)
        assert _size * n_shards == step
        cnt = 0
        # t0 = time.time()
        # if verbose:
        #     t1 = time.time()

        res = self._es.search(self._index, self._doc_type, body=q,
                              size=_size, search_type='scan', scroll=scroll, **kwargs)
        # double check initial scroll request returns no hits
        # assert len(res['hits']['hits']) == 0
        assert not res['hits']['hits']

        while 1:
            # if verbose:
            #     t1 = time.time()
            res = self._es.scroll(res['_scroll_id'], scroll=scroll)
            # if len(res['hits']['hits']) == 0:
            if not res['hits']['hits']:
                break
            else:
                for rawdoc in res['hits']['hits']:
                    if rawdoc.get('_source', False) and only_source:
                        doc = rawdoc['_source']
                        doc["_id"] = rawdoc["_id"]
                        yield doc
                    else:
                        yield rawdoc
                    cnt += 1

        assert cnt == n, "Error: scroll query terminated early [{}, {}], please retry.\nLast response:\n{}".format(cnt, n, res)

    @wrapper
    def get_id_list(self, step=None, verbose=True):
        step = step or self.step
        cur = self.doc_feeder(step=step, _source=False, verbose=verbose)
        for doc in cur:
            yield doc['_id']

    @wrapper
    def get_docs(self, ids, step=None, only_source=True, **mget_args):
        ''' Return matching docs for given ids iterable, if not found return None.
            A generator is returned to the matched docs.  If only_source is False,
            the entire document is returned, otherwise only the source is returned. '''
        # chunkify
        step = step or self.step
        for chunk in iter_n(ids, step):
            chunk_res = self._es.mget(body={"ids": chunk}, index=self._index,
                                      doc_type=self._doc_type, **mget_args)
            for rawdoc in chunk_res['docs']:
                if (('found' not in rawdoc) or (('found' in rawdoc) and not rawdoc['found'])):
                    continue
                elif not only_source:
                    yield rawdoc
                else:
                    doc = rawdoc['_source']
                    doc["_id"] = rawdoc["_id"]
                    yield doc

    def find_biggest_doc(self, fields_li, min=5, return_doc=False):     # pylint: disable=redefined-builtin
        """return the doc with the max number of fields from fields_li."""
        for n in range(len(fields_li), min - 1, -1):
            for field_set in itertools.combinations(fields_li, n):
                q = ' AND '.join(["_exists_:" + field for field in field_set])
                q = {'query': {"query_string": {"query": q}}}
                cnt = self.count(q)
                if cnt > 0:
                    if return_doc:
                        res = self._es.search(index=self._index, doc_type=self._doc_type, body=q, size=cnt)
                        return res
                    else:
                        return (cnt, q)

    def snapshot(self, repo, snapshot, mode=None, **params):
        body = {"indices": self._index}
        if mode == "purge":
            # Note: this works, just for small one when deletion is done instantly
            try:
                self._es.snapshot.get(repo, snapshot)
                # if we can get it, we have to delete it
                self._es.snapshot.delete(repo, snapshot)
            except NotFoundError:
                # ok, nothing to delete/purge
                pass
        try:
            return self._es.snapshot.create(repo, snapshot, body=body, params=params)
        except RequestError as e:
            raise IndexerException("Can't snapshot '%s': %s" % (self._index, e))

    def restore(self, repo_name, snapshot_name, index_name=None, purge=False, body=None):
        index_name = index_name or snapshot_name
        if purge:
            try:
                self._es.indices.get(index_name)
                # if we get there, it exists, delete it
                self._es.indices.delete(index_name)
            except NotFoundError:
                # no need to delete it,
                pass
        try:
            # this is just about renaming index within snapshot to index_name
            body = {
                "indices": snapshot_name,   # snaphost name is the same as index in snapshot
                "rename_replacement": index_name,
                "ignore_unavailable": True,
                "rename_pattern": "(.+)",
                "include_global_state": True
            }
            return self._es.snapshot.restore(repo_name, snapshot_name, body=body)
        except TransportError as e:
            raise IndexerException("Can't restore snapshot '%s' (does index '%s' already exist ?): %s" %
                                   (snapshot_name, index_name, e))

    def get_repository(self, repo_name):
        try:
            return self._es.snapshot.get_repository(repo_name)
        except NotFoundError:
            raise IndexerException("Repository '%s' doesn't exist" % repo_name)

    def create_repository(self, repo_name, settings):
        try:
            self._es.snapshot.create_repository(repo_name, settings)
        except TransportError as e:
            raise IndexerException("Can't create snapshot repository '%s': %s" % (repo_name, e))

    def get_snapshot_status(self, repo, snapshot):
        return self._es.snapshot.status(repo, snapshot)

    def get_restore_status(self, index_name=None):
        index_name = index_name or self._index
        recov = self._es.indices.recovery(index_name)
        if index_name not in recov:
            return {"status": "INIT", "progress": "0%"}
        shards = recov[index_name]["shards"]
        # get all shards status
        shards_status = [s["stage"] for s in shards]
        done = len([s for s in shards_status if s == "DONE"])
        if set(shards_status) == {"DONE"}:
            return {"status": "DONE", "progress": "100%"}
        else:
            return {"status": "IN_PROGRESS", "progress": "%.2f%%" % (done/len(shards_status)*100)}


class MappingError(Exception):
    pass

def generate_es_mapping(inspect_doc, init=True, level=0):
    """Generate an ES mapping according to "inspect_doc", which is
    produced by biothings.utils.inspect module"""
    map_tpl = {
        int: {"type": "integer"},
        bool: {"type": "boolean"},
        float: {"type": "float"},
        str: {"type": "keyword", "normalizer": "keyword_lowercase_normalizer"},    # not splittable (like an ID for instance)
        splitstr: {"type": "text"},
    }
    # inspect_doc, if it's been jsonified, contains keys with type as string,
    # such as "<class 'str'>". This is not a real type and we need to convert them
    # back to actual types. This is transparent if inspect_doc isalready in proper format
    pat = re.compile(r"<class '(\w+)'>")

    def str2type(k):
        if isinstance(k, str):
            mat = pat.findall(k)
            if mat:
                return eval(mat[0])     # actual type
            else:
                return k
        else:
            return k
    inspect_doc = dict_walk(inspect_doc, str2type)

    mapping = {}
    errors = []
    none_type = type(None)
    if init and "_id" not in inspect_doc:
        errors.append("No _id key found, document won't be indexed. (doc: %s)" % inspect_doc)
    for rootk in inspect_doc:
        if rootk == "_id":
            keys = list(inspect_doc[rootk].keys())
            if str in keys and splitstr in keys:
                keys.remove(str)
            if not len(keys) == 1 or (keys[0] != str and keys[0] != splitstr):
                errors.append("_id fields should all be a string type (got: %s)" % keys)
            # it was just a check, it's not part of the mapping
            continue
        if rootk == "_stats":
            continue
        if isinstance(rootk, type(None)):     # if rootk == type(None):
            # value can be null, just skip it
            continue
        # some inspect report have True as value, others have dict (will all have dict eventually)
        if inspect_doc[rootk] is True:
            inspect_doc[rootk] = {}
        keys = list(inspect_doc[rootk].keys())
        # if dict, it can be a dict containing the type (no explore needed) or a dict
        # containing more keys (explore needed)
        if list in keys:
            # we explore directly the list w/ inspect_doc[rootk][list] as param.
            # (similar to skipping list type, as there's no such list type in ES mapping)
            # carefull: there could be list of list, if which case we move further into the structure
            # to skip them
            toexplore = inspect_doc[rootk][list]
            while list in toexplore:
                toexplore = toexplore[list]
            if len(toexplore) > 1:
                # we want to make sure that, whatever the structure, the types involved were the same
                # Exception: None is allowed with other types (translates to 'null' in ES)
                # other_types = set([k for k in toexplore.keys() if k != list and isinstance(k, type) and k is not type(None)])    # TODO: Confirm this line
                other_types = {k for k in toexplore.keys() if k != list and isinstance(k, type) and not isinstance(k, none_type)}
                # some mixes are allowed by ES
                if {int, float}.issubset(other_types):
                    other_types.discard(int)    # float > int
                    toexplore.pop(int)
                if len(other_types) > 1:
                    raise Exception("Mixing types for key '%s': %s" % (rootk, other_types))
            res = generate_es_mapping(toexplore, init=False, level=level+1)
            # is it the only key or do we have more ? (ie. some docs have data as "x", some
            # others have list("x")
            # list was either a list of values (end of tree) or a list of dict. Depending
            # on that, we add "properties" (when list of dict) or not (when list of values)
            if type in set(map(type, inspect_doc[rootk][list])):
                mapping[rootk] = res
            else:
                mapping[rootk] = {"properties": {}}
                mapping[rootk]["properties"] = res
        elif set(map(type, keys)) == {type}:
            # it's a type declaration, no explore
            # typs = list(map(type, [k for k in keys if k is not type(None)]))    # TODO: Confirm this line
            typs = list(map(type, [k for k in keys if not isinstance(k, none_type)]))
            if len(typs) > 1:
                errors.append("More than one type (key:%s,types:%s)" % (repr(rootk), repr(keys)))
            try:
                typ = list(inspect_doc[rootk].keys())
                # ther can still be more than one type, if we have a None combined with
                # the "correct" one. We allow None as a combined type, but we want to ignore
                # it when we want to find the mapping
                if len(typ) == 1:
                    typ = typ[0]
                else:
                    # typ = [t for t in typ if t is not type(None)][0]      # TODO: Confirm this line
                    typ = [t for t in typ if not isinstance(t, none_type)][0]
                if typ is nan or typ is inf:
                    raise TypeError(typ)
                mapping[rootk] = map_tpl[typ]
            except KeyError:
                errors.append("Can't find map type %s for key %s" % (inspect_doc[rootk], rootk))
            except TypeError:
                errors.append("Type %s for key %s isn't allowed in ES mapping" % (typ, rootk))
        elif inspect_doc[rootk] == {}:
            typ = rootk
            return map_tpl[typ]
        else:
            mapping[rootk] = {"properties": {}}
            mapping[rootk]["properties"] = generate_es_mapping(inspect_doc[rootk], init=False, level=level+1)
    if errors:
        raise MappingError("Error while generating mapping", errors)
    return mapping


######################@#
# ES as HUB DB backend #
#@######################
from biothings.utils.hub_db import IDatabase
from biothings.utils.dotfield import parse_dot_fields
from biothings.utils.dataload import update_dict_recur
from biothings.utils.common import json_serial


def get_hub_db_conn():
    return Database()

def get_src_conn():
    return get_hub_db_conn()

def get_src_dump():
    db = Database()
    return db[db.CONFIG.DATA_SRC_DUMP_COLLECTION]

def get_src_master():
    db = Database()
    return db[db.CONFIG.DATA_SRC_MASTER_COLLECTION]

def get_src_build():
    db = Database()
    return db[db.CONFIG.DATA_SRC_BUILD_COLLECTION]

def get_src_build_config():
    db = Database()
    return db[db.CONFIG.DATA_SRC_BUILD_CONFIG_COLLECTION]

def get_data_plugin():
    db = Database()
    return db[db.CONFIG.DATA_PLUGIN_COLLECTION]

def get_api():
    db = Database()
    return db[db.CONFIG.API_COLLECTION]

def get_cmd():
    db = Database()
    return db[db.CONFIG.CMD_COLLECTION]

def get_event():
    db = Database()
    return db[db.CONFIG.EVENT_COLLECTION]

def get_hub_config():
    db = Database()
    return db[getattr(db.CONFIG, "HUB_CONFIG_COLLECTION", "hub_config")]

def get_source_fullname(col_name):
    pass

def get_last_command():
    conn = get_hub_db_conn().get_conn()
    cmd = get_cmd()
    res = conn.search(cmd.dbname, cmd.colname, {
        "query": {
            "match_all": {}
        },
        "size": 1,
        "sort": [
            {
                "_id": {
                    "order": "desc"
                }
            }
        ]
    })
    if res["hits"]["hits"]:
        return res["hits"]["hits"][0]
    else:
        return None


class Database(IDatabase):

    CONFIG = None   # will be set by bt.utils.hub_db.setup()

    def __init__(self):
        super(Database, self).__init__()
        self.name = self.CONFIG.DATA_HUB_DB_DATABASE
        self.es_host = self.CONFIG.HUB_DB_BACKEND["host"]
        self.cols = {}

    @property
    def address(self):
        return self.es_host

    def setup(self):
        pass

    def get_conn(self):
        return get_es(self.es_host)

    def create_collection(self, colname):
        return self[colname]

    def create_if_needed(self, colname):
        conn = self.get_conn()
        # add dot to make it a special index so it's hidden by default in ES gui
        idxcolname = "%s_%s" % (self.name, colname)
        # it's not usefull to scale internal hubdb
        body = {
            'settings': {
                'number_of_shards': 1,
                "number_of_replicas": 0,
            }
        }
        if not conn.indices.exists(idxcolname):
            conn.indices.create(idxcolname, body=body)
            conn.indices.put_mapping(colname, {"dynamic": True}, index=idxcolname)

    def __getitem__(self, colname):
        if colname not in self.cols:
            self.create_if_needed(colname)
            self.cols[colname] = Collection(colname, self)
        return self.cols[colname]


class Collection(object):

    def __init__(self, colname, db):
        self.colname = colname
        self.db = db

    def get_conn(self):
        return self.db.get_conn()

    @property
    def dbname(self):
        return "%s_%s" % (self.db.name, self.colname)

    @property
    def name(self):
        return self.colname

    @property
    def database(self):
        return self.db

    def find_one(self, *args, **kwargs):
        return self.find(*args, find_one=True)

    def find(self, *args, **kwargs):
        results = []
        query = {}
        # if args and len(args) == 1 and isinstance(args[0], dict) and len(args[0]) > 0:
        if args and len(args) == 1 and isinstance(args[0], dict) and args[0]:
            query = {"query": {"match": args[0]}}
        # it's key/value search, let's iterate
        res = self.get_conn().search(self.dbname, self.colname, query, size=10000)
        for _src in res["hits"]["hits"]:
            doc = {"_id": _src["_id"]}
            doc.update(_src["_source"])
            if "find_one" in kwargs:
                return doc
            else:
                results.append(doc)
        if "find_one" in kwargs:
            # we didn't find it if we get there
            return None
        return results

    def insert_one(self, doc, check_unique=True):
        assert "_id" in doc
        _id = doc.pop("_id")
        res = self.get_conn().index(self.dbname, self.colname, doc, id=_id, refresh=True)
        if check_unique and not res["result"] == "created":
            raise Exception("Couldn't insert document '%s'" % doc)

    def update_one(self, query, what, upsert=False):
        assert (len(what) == 1 and ("$set" in what or "$unset" in what or "$push" in what)), \
               "$set/$unset/$push operators not found"
        doc = self.find_one(query)
        if doc:
            if "$set" in what:
                # parse_dot_fields uses json.dumps internally, we can to make
                # sure everything is serializable first
                what = json.loads(json.dumps(what, default=json_serial))
                what = parse_dot_fields(what["$set"])
                doc = update_dict_recur(doc, what)
            elif "$unset" in what:
                for keytounset in what["$unset"].keys():
                    doc.pop(keytounset, None)
            elif "$push" in what:
                for listkey, elem in what["$push"].items():
                    assert "." not in listkey, "$push not supported for nested keys: %s" % listkey
                    doc.setdefault(listkey, []).append(elem)

            self.save(doc)
        elif upsert:
            assert "_id" in query, "Can't upsert without _id"
            assert "$set" in what, "Upsert needs $set operator (it makes sense...)"
            doc = what["$set"]
            doc["_id"] = query["_id"]
            self.save(doc)

    def update(self, query, what):
        docs = self.find(query)
        for doc in docs:
            self.update_one({"_id": doc["_id"]}, what)

    def save(self, doc):
        return self.insert_one(doc, check_unique=False)

    def replace_one(self, query, doc, *args, **kwargs):
        orig = self.find_one(query)
        if orig:
            self.insert_one(doc, check_unique=False)

    def remove(self, query):
        docs = self.find(query)
        conn = self.get_conn()
        for doc in docs:
            conn.delete(self.dbname, self.colname, id=doc["_id"], refresh=True)

    def count(self):
        return self.get_conn().count(self.dbname, self.colname)["count"]

    def __getitem__(self, _id):
        return self.find_one({"_id": _id})



import logging
import mimetypes
import os
from urllib.parse import urlparse

from boto import connect_s3
import boto3
import botocore.exceptions

try:
    from biothings import config
except ImportError:
    # assuming key, secret and bucket will be passed
    # to all functions
    pass


def key_exists(bucket, s3key, aws_key=None, aws_secret=None):
    client = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    try:
        client.head_object(Bucket=bucket, Key=s3key)
        return True
    except Exception as e:
        if e.response["Error"]["Code"] == "404":
            return False
        else:
            raise


def send_s3_file(localfile, s3key, overwrite=False, permissions=None, metadata=None,
                 content=None, content_type=None, aws_key=None, aws_secret=None, s3_bucket=None):
    '''save a localfile to s3 bucket with the given key.
       bucket is set via S3_BUCKET
       it also save localfile's lastmodified time in s3 file's metadata
    '''
    metadata = metadata or {}
    try:
        aws_key = aws_key or getattr(config, "AWS_SECRET")
        aws_secret = aws_secret or getattr(config, "AWS_SECRET")
        s3_bucket = s3_bucket or getattr(config, "S3_BUCKET")
    except AttributeError:
        logging.info("Skip sending file to S3, missing information in config file: AWS_KEY, AWS_SECRET or S3_BUCKET")
        return
    s3 = connect_s3(aws_key, aws_secret)
    bucket = s3.get_bucket(s3_bucket)
    # bucket_location = bucket.get_location()  # TODO: delete this unused variable
    k = bucket.new_key(s3key)
    if not overwrite:
        assert not k.exists(), 's3key "{}" already exists.'.format(s3key)
    for header in metadata:
        k.set_metadata(header, metadata[header])
    if content_type:
        k.content_type = content_type
    if content:
        k.set_contents_from_string(content)
    else:
        assert os.path.exists(localfile), 'localfile "{}" does not exist.'.format(localfile)
        lastmodified = os.stat(localfile)[-2]
        k.set_metadata('lastmodified', lastmodified)
        k.set_contents_from_filename(localfile)
    if permissions:
        k.set_acl(permissions)


def send_s3_big_file(localfile, s3key, overwrite=False, acl=None,
                     aws_key=None, aws_secret=None, s3_bucket=None,
                     storage_class=None):
    """
    Multiparts upload for file bigger than 5GiB
    """
    # TODO: maybe merge with send_s3_file() based in file size ? It would need boto3 migration
    try:
        aws_key = aws_key or getattr(config, "AWS_SECRET")
        aws_secret = aws_secret or getattr(config, "AWS_SECRET")
        s3_bucket = s3_bucket or getattr(config, "S3_BUCKET")
    except AttributeError:
        logging.info("Skip sending file to S3, missing information in config file: AWS_KEY, AWS_SECRET or S3_BUCKET")
        return
    client = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    if not overwrite and key_exists(s3_bucket, s3key, aws_key, aws_secret):
        raise Exception("Key '%s' already exist" % s3key)
    tfr_config = boto3.s3.transfer.TransferConfig(
        multipart_threshold=1024 * 25,
        max_concurrency=10,
        multipart_chunksize=1024 * 25,
        use_threads=True
    )
    extra = {
        "ACL": acl or "private",
        "ContentType": mimetypes.MimeTypes().guess_type(localfile)[0] or "binary/octet-stream",
        "StorageClass": storage_class or "REDUCED_REDUNDANCY"
    }
    client.upload_file(Filename=localfile, Bucket=s3_bucket, Key=s3key, ExtraArgs=extra, Config=tfr_config)


def get_s3_file(s3key, localfile=None, return_what=False,
                aws_key=None, aws_secret=None, s3_bucket=None):
    aws_key = aws_key or getattr(config, "AWS_SECRET")
    aws_secret = aws_secret or getattr(config, "AWS_SECRET")
    s3_bucket = s3_bucket or getattr(config, "S3_BUCKET")
    localfile = localfile or os.path.basename(s3key)
    s3 = connect_s3(aws_key, aws_secret)
    bucket = s3.get_bucket(s3_bucket)
    k = bucket.get_key(s3key)
    if not k:
        raise FileNotFoundError(s3key)
    if return_what == "content":
        return k.get_contents_as_string()
    elif return_what == "key":
        return k
    else:
        k.get_contents_to_filename(localfile)


def get_s3_folder(s3folder, basedir=None, aws_key=None, aws_secret=None, s3_bucket=None):
    aws_key = aws_key or getattr(config, "AWS_SECRET")
    aws_secret = aws_secret or getattr(config, "AWS_SECRET")
    s3_bucket = s3_bucket or getattr(config, "S3_BUCKET")
    s3 = connect_s3(aws_key, aws_secret)
    bucket = s3.get_bucket(s3_bucket)
    cwd = os.getcwd()
    try:
        if basedir:
            os.chdir(basedir)
        if not os.path.exists(s3folder):
            os.makedirs(s3folder)
        for k in bucket.list(s3folder):
            get_s3_file(k.key, localfile=k.key, aws_key=aws_key, aws_secret=aws_secret, s3_bucket=s3_bucket)
    finally:
        os.chdir(cwd)


def send_s3_folder(folder, s3basedir=None, acl=None, overwrite=False,
                   aws_key=None, aws_secret=None, s3_bucket=None):
    aws_key = aws_key or getattr(config, "AWS_SECRET")
    aws_secret = aws_secret or getattr(config, "AWS_SECRET")
    s3_bucket = s3_bucket or getattr(config, "S3_BUCKET")
    s3 = connect_s3(aws_key, aws_secret)
    s3.get_bucket(s3_bucket)    # check if s3_bucket exists
    cwd = os.getcwd()
    if not s3basedir:
        s3basedir = os.path.basename(cwd)
    for localf in [f for f in os.listdir(folder) if not f.startswith(".")]:
        fullpath = os.path.join(folder, localf)
        if os.path.isdir(fullpath):
            send_s3_folder(fullpath, os.path.join(s3basedir, localf),
                           overwrite=overwrite, acl=acl,
                           aws_key=aws_key, aws_secret=aws_secret, s3_bucket=s3_bucket)
        else:
            send_s3_big_file(fullpath, os.path.join(s3basedir, localf),
                             overwrite=overwrite, acl=acl,
                             aws_key=aws_key, aws_secret=aws_secret, s3_bucket=s3_bucket)


def get_s3_url(s3key, aws_key=None, aws_secret=None, s3_bucket=None):
    try:
        k = get_s3_file(s3key, return_what="key",
                        aws_key=aws_key, aws_secret=aws_secret, s3_bucket=s3_bucket)
    except FileNotFoundError:
        return None
    # generate_url will include some acdesskey, signature, etc... we want to remove this
    # as the bucket is public anyway and want "clean" url
    url = k.generate_url(expires_in=0)  # never (and whatever, we
    return urlparse(url)._replace(query="").geturl()


def create_bucket(name, region=None, aws_key=None, aws_secret=None, acl=None,
                  ignore_already_exists=False):
    """Create a S3 bucket "name" in optional "region". If aws_key and aws_secret
    are set, S3 client will these, otherwise it'll use default system-wide setting.
    "acl" defines permissions on the bucket: "private" (default), "public-read",
    "public-read-write" and "authenticated-read"
    """
    client = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    acl = acl or "private"
    kwargs = {"ACL": acl, "Bucket": name}
    if region:
        kwargs["CreateBucketConfiguration"] = {"LocationConstraint": region}
    try:
        client.create_bucket(**kwargs)
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "BucketAlreadyOwnedByYou" and not ignore_already_exists:
            raise

def set_static_website(name, aws_key=None, aws_secret=None, index="index.html", error="error.html"):
    client = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    conf = {'IndexDocument': {'Suffix': index},
            'ErrorDocument': {'Key': error}}
    client.put_bucket_website(Bucket=name, WebsiteConfiguration=conf)
    location = client.get_bucket_location(Bucket=name)
    region = location["LocationConstraint"]
    # generate website URL
    return "http://%(name)s.s3-website-%(region)s.amazonaws.com" % {"name": name, "region": region}



import logging

from elasticsearch_async.connection import AIOHttpConnection

logger = logging.getLogger("elasticsearch")
tracer = logging.getLogger("elasticsearch.trace")

class BiothingsAIOHttpConnection(AIOHttpConnection):

    def _log_trace(self, method, path, body, status_code, response, duration):
        if not tracer.isEnabledFor(logging.INFO) or not tracer.handlers:
            return

        # include pretty in trace curls
        path = path.replace("?", "?pretty&", 1) if "?" in path else path + "?pretty"
        if self.url_prefix:
            path = path.replace(self.url_prefix, "", 1)
        tracer.info(
            "curl %s-X%s '%s' -d '%s'",  # override: path is full url
            "-H 'Content-Type: application/json' " if body else "",
            method,
            path,
            self._pretty_json(body) if body else "",
        )

        if tracer.isEnabledFor(logging.DEBUG):
            # override
            if response:
                tracer.debug(
                    "#[%s] (%.3fs)\n#%s",
                    status_code,
                    duration,
                    self._pretty_json(response).replace("\n", "\n#") if response else "",
                )
            else:  # remove empty body output
                tracer.debug(
                    "#[%s] (%.3fs)",
                    status_code,
                    duration
                )

    def log_request_success(
        self, method, full_url, path, body, status_code, response, duration
    ):
        """ Log a successful API call.  """

        # body has already been serialized to utf-8, deserialize it for logging
        if body:
            try:
                body = body.decode("utf-8", "ignore")
            except AttributeError:
                pass

        logger.info(
            "%s %s [status:%s request:%.3fs]", method, full_url, status_code, duration
        )
        logger.debug("> %s", body)
        logger.debug("< %s", response)

        # override, pass in full url instead of path
        self._log_trace(method, full_url, body, status_code, response, duration)

    def log_request_fail(
        self,
        method,
        full_url,
        path,
        body,
        duration,
        status_code=None,
        response=None,
        exception=None,
    ):
        """ Log an unsuccessful API call.  """
        # do not log 404s on HEAD requests
        if method == "HEAD" and status_code == 404:
            return
        logger.warning(
            "%s %s [status:%s request:%.3fs]",
            method,
            full_url,
            status_code or "N/A",
            duration,
            # Override
            # exc_info=exception is not None,
        )

        # body has already been serialized to utf-8, deserialize it for logging
        # TODO: find a better way to avoid (de)encoding the body back and forth
        if body:
            try:
                body = body.decode("utf-8", "ignore")
            except AttributeError:
                pass

        logger.debug("> %s", body)

        # override, pass in full url instead of path
        self._log_trace(method, full_url, body, status_code, response, duration)



from biothings.hub.datatransform.datatransform import DataTransform  # noqa: F401
from biothings.hub.datatransform.datatransform import IDStruct  # noqa: F401
from biothings.hub.datatransform.datatransform import DataTransformEdge  # noqa: F401
from biothings.hub.datatransform.datatransform import RegExEdge  # noqa: F401
from biothings.hub.datatransform.datatransform import nested_lookup  # noqa: F401

# only for testing purpose, this one is not optimized
#from biothings.hub.datatransform.datatransform_serial import DataTransformSerial

from biothings.hub.datatransform.datatransform_mdb import DataTransformMDB  # noqa: F401
from biothings.hub.datatransform.datatransform_mdb import MongoDBEdge  # noqa: F401
from biothings.hub.datatransform.datatransform_mdb import CIMongoDBEdge  # noqa: F401

# case-insensitive IDStruct class
from biothings.hub.datatransform.ciidstruct import CIIDStruct  # noqa: F401

# this involved biothings client dependency, skip it for now
from biothings.hub.datatransform.datatransform_api import MyChemInfoEdge  # noqa: F401
from biothings.hub.datatransform.datatransform_api import MyGeneInfoEdge  # noqa: F401



import logging
import asyncio
import json

from biothings.utils.hub_db import ChangeListener

import sockjs.tornado


class WebSocketConnection(sockjs.tornado.SockJSConnection):
    """
    Listen to Hub DB through a listener object, and publish
    events to any client connected
    """

    clients = set()

    def __init__(self, session, listeners):
        """
        SockJSConnection.__init__() takes only a session as argument, and there's
        no way to pass custom settings. In order to use that class, we need to use partial
        to partially init the instance with 'listeners' and let the rest use the 'session'

        parameter:
          pconn = partial(WebSocketConnection,listeners=listeners)
          ws_router = sockjs.tornado.SockJSRouter(pconn,"/path")
        """
        if type(listeners) != list:
            listeners = [listeners]
        self.listeners = listeners
        # propagate connection so listeners can access it and trigger message sending
        for listener in self.listeners:
            listener.socket = self
        super(WebSocketConnection, self).__init__(session)

    def publish(self, message):
        self.broadcast(self.__class__.clients, message)

    def on_open(self, info):
        # Send that someone joined
        self.broadcast(self.__class__.clients, "Someone joined. %s" % info)
        # Add client to the clients list
        self.__class__.clients.add(self)

    def on_message(self, message):
        err = None
        strerr = None
        try:
            message = json.loads(message)
            if message["op"] == "ping":
                self.send({"op": "pong"})
        except json.JSONDecodeError:
            strerr = "malformed json message: %s" % message
            err = json.JSONDecodeError(strerr)
        except KeyError:
            strerr = "malformed socket message: %s" % message
            err = KeyError(strerr)
        except Exception as e:
            strerr = "Unable to process message '%s': %s" % (message, e)
            err = Exception(strerr)
        if err:
            self.send({"error": strerr})
            raise err

    def on_close(self):
        # Remove client from the clients list and broadcast leave message
        self.__class__.clients.remove(self)
        self.broadcast(self.__class__.clients, "Someone left.")


class HubDBListener(ChangeListener):
    """
    Get events from Hub DB and propagate them through the
    websocket instance
    """
    def read(self, event):
        # self.socket is set while initalizing the websocket connection
        self.socket.publish(event)


class LogListener(ChangeListener):
    # IMPORTANT: no logging calls here, or infinite loop
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.socket = None

    def read(self, event):
        if self.socket:
            # make sure there's a loop in current thread
            try:
                logging.disable(logging.CRITICAL)
                loop = asyncio.get_event_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            finally:
                logging.disable(logging.NOTSET)

            try:
                self.socket.publish(event)
            except Exception as e:
                # can't log anything there, but we don't want a problem with
                # issuing the log statements through the websocket to cause
                # any error in the caller
                print(e)
                pass


class ShellListener(LogListener):
    pass



''' web utils go here. '''
# randos
def sum_arg_dicts(*arg, **kwarg):
    ''' any dicts passed to the args of this function (either positional or keyword) are combined (using update) '''
    _ret = {}
    for _val in arg + tuple(kwarg.values()):
        if isinstance(_val, dict):
            _ret.update(_val)
    return _ret
